"""This package includes circuit minimizer implementations."""
from __future__ import annotations

from bqskit.ir.opt.minimizers.ceres import CeresMinimizer
from bqskit.ir.opt.minimizers.lbfgs import LBFGSMinimizer
from bqskit.ir.opt.minimizers.scipy import ScipyMinimizer

__all__ = ['LBFGSMinimizer', 'CeresMinimizer', 'ScipyMinimizer']



"""This module implements the BGate."""
from __future__ import annotations

from numpy import pi
from scipy.linalg import expm

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.pauli import PauliMatrices
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class BGate(ConstantGate, QubitGate):
    """
    The 2 qubit B gate.

    The B gate is given by the following unitary:

    .. math::
        \\exp(i * \\pi/4 * \\sigma_{xx}) * \\exp(i * \\pi/8 * \\sigma_{yy})

    References:
        - https://arxiv.org/pdf/quant-ph/0312193.pdf
    """

    _num_qudits = 2
    _qasm_name = 'b'

    def __init__(self) -> None:
        """Construct a BGate."""
        paulis = PauliMatrices(2)
        xx = paulis[5]
        yy = paulis[10]
        mat = expm(1j * pi / 4 * xx) @ expm(1j * pi / 8 * yy)
        self._utry = UnitaryMatrix(mat)



"""This module implements many helper functions to check types."""
from __future__ import annotations

import logging
import numbers
from collections.abc import Sequence
from typing import Any
from typing import Iterable
from typing import Mapping
from typing import Sized

import numpy as np
import numpy.typing as npt
from typing_extensions import TypeGuard


_logger = logging.getLogger(__name__)


def is_iterable(x: Any) -> TypeGuard[Iterable[Any]]:
    """Return true if x is an iterable object."""
    try:
        iterator = iter(x)  # noqa: F841
        return True
    except TypeError:
        return False


def is_sized(x: Any) -> TypeGuard[Sized]:
    """Return true if x is a sized object."""
    try:
        length = len(x)  # noqa: F841
        return True
    except TypeError:
        return False


def is_sequence(x: Any) -> TypeGuard[Sequence[Any]]:
    """Return true if x is a sequence."""
    return isinstance(x, (Sequence, np.ndarray))


def is_mapping(x: Any) -> TypeGuard[Mapping[Any, Any]]:
    """Return true if x is a mapping."""
    return isinstance(x, Mapping)


def is_numeric(x: Any) -> TypeGuard[numbers.Number]:
    """Return true if x is numeric."""
    return isinstance(x, numbers.Number) and not isinstance(x, bool)


def is_complex(x: Any) -> TypeGuard[complex]:
    """Return true if x is complex."""
    return isinstance(x, numbers.Complex)


def is_real_number(x: Any) -> TypeGuard[float]:
    """Return true if `x` is a real number."""
    return isinstance(x, numbers.Real)


def is_integer(x: Any) -> TypeGuard[int]:
    """Return true if x is an integer."""
    return isinstance(x, (int, np.int_))


def is_bool(x: Any) -> TypeGuard[bool]:
    """Return true if x is a boolean value."""
    return isinstance(x, (bool, np.bool_))


def is_sequence_of_int(x: Any) -> TypeGuard[Sequence[int]]:
    """Return true if x is a sequence of integers."""
    return is_sequence(x) and all([is_integer(xi) for xi in x])


def is_valid_radixes(
    radixes: Sequence[int],
    num_qudits: int | None = None,
) -> bool:
    """
    Determine if the sequence of radixes is valid. Radixes must be integers
    greater than or equal to 2. If num_qudits is specified, then the length of
    radixes must be equal to num_qudits.

    Args:
        radixes (Sequence[int]): The radixes to check.

        num_qudits (int | None): The total number of qudits.
            Length of `radixes` should be equal to this. If None,
            don't check.

    Returns:
        bool: True if the radixes are valid.
    """

    if not is_sequence(radixes):
        return False

    if not all([is_integer(qudit) for qudit in radixes]):
        fail_idx = [is_integer(qudit) for qudit in radixes].index(False)
        _logger.debug(
            'Radixes is not a tuple of ints, got: %s.' % type(
                radixes[fail_idx],
            ),
        )
        return False

    if not all([radix >= 2 for radix in radixes]):
        _logger.debug('Radixes invalid; radix indices must be >= 2.')
        return False

    if num_qudits is not None and len(radixes) != num_qudits:
        _logger.debug('Invalid number of radixes.')
        return False

    return True


def is_vector(V: np.typing.ArrayLike) -> bool:
    """Return true if V is a vector."""

    if not isinstance(V, np.ndarray):
        V = np.array(V)

    if len(V.shape) != 1:
        _logger.debug('V is not an 1-dimensional array.')
        return False

    if V.dtype.kind not in 'biufc':
        _logger.debug('V is not a numeric array.')
        return False

    return True


def is_matrix(M: np.typing.ArrayLike) -> bool:
    """Return true if M is a matrix."""

    if not isinstance(M, np.ndarray):
        M = np.array(M)

    if len(M.shape) != 2:
        # _logger.debug('M is not an 2-dimensional array.')
        return False

    if M.dtype.kind not in 'biufc':
        # _logger.debug('M is not a numeric array.')
        return False

    return True


def is_square_matrix(M: np.typing.ArrayLike) -> bool:
    """Return true if M is a square matrix."""

    if not isinstance(M, np.ndarray):
        M = np.array(M)

    if not is_matrix(M):
        return False

    if M.shape[0] != M.shape[1]:
        return False

    return True


def is_hermitian(H: npt.NDArray[np.complex128], tol: float = 1e-8) -> bool:
    """Return true if H is a hermitian matrix."""

    if not is_square_matrix(H):
        return False

    if not np.allclose(H, H.conj().T, rtol=0, atol=tol):
        if _logger.isEnabledFor(logging.DEBUG):
            norm = np.linalg.norm(H - H.conj().T)
            _logger.debug(
                'Failed hermitian condition, ||H - H^d|| = %e'
                % norm,
            )
        return False

    return True


def is_skew_hermitian(H: npt.NDArray[np.complex128], tol: float = 1e-8) -> bool:
    """Return true if H is a skew hermitian matrix."""

    if not is_square_matrix(H):
        return False

    if not np.allclose(-H, H.conj().T, rtol=0, atol=tol):
        if _logger.isEnabledFor(logging.DEBUG):
            norm = np.linalg.norm(-H - H.conj().T)
            _logger.debug(
                'Failed skew hermitian condition, ||H - H^d|| = %e'
                % norm,
            )
        return False

    return True



"""This module implements the ConstantUnitaryGate."""
from __future__ import annotations

from typing import Sequence

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.qis.unitary.unitarymatrix import UnitaryLike
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class ConstantUnitaryGate(ConstantGate):
    """An arbitrary constant unitary operator."""

    def __init__(
        self,
        utry: UnitaryLike,
        radixes: Sequence[int] = [],
    ) -> None:
        """
        Construct a constant unitary operator.

        Args:
            utry (UnitaryLike): The operation as a unitary matrix.

            radixes (Sequence[int]): The number of orthogonal states
                for each qudit this gate will act on. Defaults to qubits.
        """
        self._utry = UnitaryMatrix(utry, radixes)
        self._num_qudits = self._utry.num_qudits
        self._radixes = self._utry.radixes

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, ConstantUnitaryGate)
            and self._utry == other._utry
        )

    def __hash__(self) -> int:
        return hash(self._utry)



"""This module implements the U3Decomposition."""
from __future__ import annotations

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.parameterized.u3 import U3Gate


class U3Decomposition(BasePass):
    """
    The U3Decomposition class.

    Convert a single-qubit circuit to U3 gate.
    """

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        if circuit.num_qudits != 1:
            raise ValueError(
                'Cannot convert multi-qudit circuit into U3 gate.',
            )

        if circuit.radixes[0] != 2:
            raise ValueError(
                'Cannot convert non-qubit circuit into U3 gate.',
            )

        utry = circuit.get_unitary()
        new_circuit = Circuit(1)
        new_circuit.append_gate(U3Gate(), 0, U3Gate().calc_params(utry))
        circuit.become(new_circuit)



"""This module implements the NodeBase abstract class."""
from __future__ import annotations

import abc
import functools
import logging
import os
import random
import selectors
import signal
import socket
import sys
import time
import traceback
from multiprocessing import Process
from multiprocessing.connection import Client
from multiprocessing.connection import Connection
from multiprocessing.connection import Listener
from queue import Queue
from threading import Thread
from types import FrameType
from typing import Any
from typing import cast
from typing import Sequence

from bqskit.runtime import default_manager_port
from bqskit.runtime import default_worker_port
from bqskit.runtime.address import RuntimeAddress
from bqskit.runtime.direction import MessageDirection
from bqskit.runtime.message import RuntimeMessage
from bqskit.runtime.result import RuntimeResult
from bqskit.runtime.task import RuntimeTask
from bqskit.runtime.worker import start_worker


class RuntimeEmployee:
    """Data structure for a boss's view of an employee."""

    def __init__(
        self,
        conn: Connection,
        total_workers: int,
        process: Process | None = None,
        num_tasks: int = 0,
    ) -> None:
        """Construct an employee with all resources idle."""
        self.conn: Connection = conn
        self.total_workers = total_workers
        self.process = process
        self.num_tasks = num_tasks
        self.num_idle_workers = total_workers

    def shutdown(self) -> None:
        """Shutdown the employee."""
        try:
            self.conn.send((RuntimeMessage.SHUTDOWN, None))
        except Exception:
            pass

        if self.process is not None:
            self.process.join()

        self.process = None
        self.conn.close()

    @property
    def has_idle_resources(self) -> bool:
        return self.num_idle_workers > 0


def send_outgoing(node: ServerBase) -> None:
    """Outgoing thread forwards messages as they are created."""
    while True:
        outgoing = node.outgoing.get()

        if not node.running:
            # NodeBase's handle_shutdown will put a dummy value in the
            # queue to wake the thread up so it can exit safely.
            # Hence the node.running check now rather than in the
            # while condition.
            break

        outgoing[0].send((outgoing[1], outgoing[2]))
        node.logger.debug(f'Sent message {outgoing[1].name}.')
        node.logger.log(1, f'{outgoing[2]}\n')


def sigint_handler(signum: int, _: FrameType | None, node: ServerBase) -> None:
    """Interrupt the node."""
    if not node.running:
        return

    node.running = False
    node.terminate_hotline.send(b'\0')
    node.logger.info('Server interrupted.')


class ServerBase:
    """Base class for all non-worker process nodes in the BQSKit Runtime."""

    def __init__(self) -> None:
        """Initialize a runtime node component."""

        self.lower_id_bound = 0
        self.upper_id_bound = int(2 ** 30)
        """
        The node starts with an ID range from 0 -> 2^30. ID ranges are then
        assigned to managers by evenly splitting this range.

        Managers then recursively split their range when connecting the sub-
        managers. Finally, workers are assigned specific ids from within this
        range.
        """

        self.running = True
        """True while the node is running."""

        self.sel = selectors.DefaultSelector()
        """Used to efficiently idle and wake when communication is ready."""

        p, self.terminate_hotline = socket.socketpair()
        self.sel.register(p, selectors.EVENT_READ, MessageDirection.SIGNAL)
        """Terminate hotline is used to unblock select while running."""

        self.logger = logging.getLogger('bqskit-runtime')
        """Logger used to print operational log messages."""

        self.employees: list[RuntimeEmployee] = []
        """Tracks this node's employees, which are managers or workers."""

        self.conn_to_employee_dict: dict[Connection, RuntimeEmployee] = {}
        """Used to find the employee associated with a message."""

        # Safely and immediately exit on interrupt signals
        handle = functools.partial(sigint_handler, node=self)
        signal.signal(signal.SIGINT, handle)

        # Start outgoing thread
        self.outgoing: Queue[tuple[Connection, RuntimeMessage, Any]] = Queue()
        self.outgoing_thread = Thread(target=send_outgoing, args=(self,))
        self.outgoing_thread.start()
        self.logger.info('Started outgoing thread.')

    def connect_to_managers(self, ipports: Sequence[tuple[str, int]]) -> None:
        """Connect to all managers given by endpoints in `ipports`."""
        d = len(ipports)
        self.step_size = (self.upper_id_bound - self.lower_id_bound) // d

        # Establish connections with all managers
        manager_conns = []
        for i, (ip, port) in enumerate(ipports):
            lb = self.lower_id_bound + (i * self.step_size)
            ub = min(
                self.lower_id_bound + ((i + 1) * self.step_size),
                self.upper_id_bound,
            )
            manager_conns.append(self.connect_to_manager(ip, port, lb, ub))
            self.logger.info(f'Connected to manager {i} at {ip}:{port}.')
            self.logger.debug(f'Gave bounds {lb=} and {ub=} to manager {i}.')

        # Wait for started messages from all managers and register them
        self.total_workers = 0
        for i, conn in enumerate(manager_conns):
            msg, num_workers = conn.recv()
            assert msg == RuntimeMessage.STARTED
            self.employees.append(RuntimeEmployee(conn, num_workers))
            self.conn_to_employee_dict[conn] = self.employees[-1]
            self.sel.register(
                conn,
                selectors.EVENT_READ,
                MessageDirection.BELOW,
            )
            self.logger.info(f'Registered manager {i} with {num_workers=}.')
            self.total_workers += num_workers
        self.num_idle_workers = self.total_workers

        self.logger.info(f'Node has {self.total_workers} total workers.')

    def connect_to_manager(
        self,
        ip: str,
        port: int,
        lb: int,
        ub: int,
    ) -> Connection:
        """
        Connect to a manager at the endpoint given by `ip` and `port`.

        Args:
            ip (str): The IP address where the manager is expected to be
                listening.

            port (int): The port number on which the manager is expected
                to be listening.

            lb (int): The ID lower bound to send to the manager.

            ub (int): The ID upper bound to send to the manager.
        """
        max_retries = 5
        wait_time = .25

        for _ in range(max_retries):
            try:
                conn = Client((ip, port))
            except ConnectionRefusedError:
                time.sleep(wait_time)
                wait_time *= 2
            else:
                conn.send((RuntimeMessage.CONNECT, (lb, ub)))
                return conn

        raise RuntimeError(f'Manager connection refused at {ip}:{port}')

    def spawn_workers(
        self,
        num_workers: int = -1,
        port: int = default_worker_port,
    ) -> None:
        """
        Spawn worker processes.

        Args:
            num_workers (int): The number of workers to spawn. If -1,
                then spawn as many workers as CPUs on the system.
                (Default: -1).

            port (int): The port this server will listen for workers on.
                Default can be found in the
                :obj:`~bqskit.runtime.default_worker_port` global variable.
        """
        if num_workers == -1:
            oscount = os.cpu_count()
            num_workers = oscount if oscount else 1

        if self.lower_id_bound + num_workers >= self.upper_id_bound:
            raise RuntimeError('Insufficient id range for workers.')

        # Create and start all worker processes
        procs = {}
        for i in range(num_workers):
            w_id = self.lower_id_bound + i
            procs[w_id] = Process(target=start_worker, args=(w_id, port))
            procs[w_id].start()
            self.logger.debug(f'Stated worker process {i}.')

        # Listen for the worker connections
        family = 'AF_INET' if sys.platform == 'win32' else None
        listener = Listener(('localhost', port), family, backlog=num_workers)
        conns = [listener.accept() for _ in range(num_workers)]
        listener.close()

        # Organize all workers into the employees data structure
        temp_reorder = {}
        for i, conn in enumerate(conns):
            msg, w_id = conn.recv()
            assert msg == RuntimeMessage.STARTED
            employee = RuntimeEmployee(conn, 1, procs[w_id])
            temp_reorder[w_id - self.lower_id_bound] = employee
            self.conn_to_employee_dict[conn] = employee

        # The employess list needs to be sorted according to the IDs
        for i in range(num_workers):
            self.employees.append(temp_reorder[i])

        # Register employee communication
        for i, employee in enumerate(self.employees):
            self.sel.register(
                employee.conn,
                selectors.EVENT_READ,
                MessageDirection.BELOW,
            )
            self.logger.info(f'Registered worker {i}.')

        self.step_size = 1
        self.total_workers = num_workers
        self.num_idle_workers = num_workers
        self.logger.info(f'Node has spawned {num_workers} workers.')

    def connect_to_workers(
        self,
        num_workers: int = -1,
        port: int = default_worker_port,
    ) -> None:
        """
        Connect to worker processes.

        Args:
            num_workers (int): The number of workers to expect. If -1,
                then expect as many workers as CPUs on the system.
                (Default: -1).

            port (int): The port this server will listen for workers on.
                Default can be found in the
                :obj:`~bqskit.runtime.default_worker_port` global variable.
        """
        if num_workers == -1:
            oscount = os.cpu_count()
            num_workers = oscount if oscount else 1

        self.logger.info(f'Expecting {num_workers} worker connections.')

        if self.lower_id_bound + num_workers >= self.upper_id_bound:
            raise RuntimeError('Insufficient id range for workers.')

        # Listen for the worker connections
        family = 'AF_INET' if sys.platform == 'win32' else None
        listener = Listener(('localhost', port), family, backlog=num_workers)
        conns = [listener.accept() for _ in range(num_workers)]
        listener.close()

        for i, conn in enumerate(conns):
            w_id = self.lower_id_bound + i
            self.outgoing.put((conn, RuntimeMessage.STARTED, w_id))
            employee = RuntimeEmployee(conn, 1)
            self.employees.append(employee)
            self.conn_to_employee_dict[conn] = employee

        # Register employee communication
        for i, employee in enumerate(self.employees):
            w_id = self.lower_id_bound + i
            assert employee.conn.recv() == (RuntimeMessage.STARTED, w_id)
            self.sel.register(
                employee.conn,
                selectors.EVENT_READ,
                MessageDirection.BELOW,
            )
            self.logger.info(f'Registered worker {i}.')

        self.step_size = 1
        self.total_workers = num_workers
        self.num_idle_workers = num_workers
        self.logger.info(f'Node has connected to {num_workers} workers.')

    def listen_once(self, ip: str, port: int) -> Connection:
        """Listen on `ip`:`port` for a connection and return on first one."""
        family = 'AF_INET' if sys.platform == 'win32' else None
        listener = Listener((ip, port), family)
        conn = listener.accept()
        listener.close()
        return conn

    def run(self) -> None:
        """Main loop."""
        self.logger.info(f'{self.__class__.__name__} running...')

        try:
            while self.running:
                # Wait for messages
                events = self.sel.select()  # Say that 5 times fast

                for key, _ in events:
                    # Unpack message, payload, and direction.
                    conn = cast(Connection, key.fileobj)
                    direction = cast(MessageDirection, key.data)

                    # If interrupted by signal, shutdown and exit
                    if direction == MessageDirection.SIGNAL:
                        self.logger.debug('Received interrupt signal.')
                        self.handle_shutdown()
                        return

                    # Unpack and Log message
                    try:
                        msg, payload = conn.recv()
                    except (EOFError, ConnectionResetError):
                        self.handle_disconnect(conn)
                        continue
                    log = f'Received message {msg.name} from {direction.name}.'
                    self.logger.debug(log)
                    self.logger.log(1, f'{payload}\n')

                    # Handle message
                    self.handle_message(msg, direction, conn, payload)

        except Exception:
            exc_info = sys.exc_info()
            error_str = ''.join(traceback.format_exception(*exc_info))
            self.logger.error(error_str)
            self.handle_system_error(error_str)

        finally:
            self.handle_shutdown()

    @abc.abstractmethod
    def handle_message(
        self,
        msg: RuntimeMessage,
        direction: MessageDirection,
        conn: Connection,
        payload: Any,
    ) -> None:
        """
        Process the message coming from `direction`.

        Args:
            msg (RuntimeMessage): The message type to handle.

            direction (MessageDirection): The direction the message came from.

            conn (Connection): The connection object where this came from.

            payload (Any): The message data.
        """

    @abc.abstractmethod
    def handle_system_error(self, error_str: str) -> None:
        """
        Handle an error in runtime code as opposed to client code.

        This is called when an error arises in runtime code not in a
        RuntimeTask's coroutine code.
        """

    def handle_shutdown(self) -> None:
        """Shutdown the node and release resources."""
        # Stop running
        self.logger.info('Shutting down node.')
        self.running = False

        # Instruct employees to shutdown
        for employee in self.employees:
            employee.shutdown()
        self.employees.clear()
        self.logger.debug('Shutdown employees.')

        # Close selector
        self.sel.close()
        self.logger.debug('Cleared selector.')

        # Close outgoing thread
        if self.outgoing_thread.is_alive():
            self.outgoing.put(b'\0')  # type: ignore
            self.outgoing_thread.join()
            self.logger.debug('Joined outgoing thread.')
            assert not self.outgoing_thread.is_alive()

    def handle_disconnect(self, conn: Connection) -> None:
        """Remove `conn` from the server."""
        self.sel.unregister(conn)
        conn.close()

        # If one of my employees crashed/shutdown/disconnected, I shutdown
        if conn in self.conn_to_employee_dict:
            self.handle_shutdown()

    def __del__(self) -> None:
        """Ensure resources are cleaned up."""
        self.handle_shutdown()

    def assign_tasks(
        self,
        tasks: Sequence[RuntimeTask],
    ) -> list[list[RuntimeTask]]:
        """
        Go through the tasks and assign each one to an employee.

        Strategy:
            - Assign first to idle workers evenly and randomly
            - Employees with more idle workers get prioritized
            - Assign remaining tasks to employees with the fewest tasks
        """
        # assignments will contain the assigned task list for each employee
        assignments: list[list[RuntimeTask]] = [[] for _ in self.employees]

        # Every employee's id repeated as many time as it has idle workers:
        idle_id_repeated_list: list[int] = sum(
            (
                [i] * e.num_idle_workers
                for i, e in enumerate(self.employees)
            ), [],
        )

        # Shuffle to reduce chance of inefficiency described in handle_waiting
        random.shuffle(idle_id_repeated_list)

        # Assign tasks to idle workers in random order
        for idle_employee_id, task in zip(idle_id_repeated_list, tasks):
            assignments[idle_employee_id].append(task)

        # Check if there are more tasks to be assigned
        num_remaining_tasks = len(tasks) - len(idle_id_repeated_list)

        if num_remaining_tasks <= 0:
            return assignments

        remaining_tasks = list(tasks[-num_remaining_tasks:])

        # Sort the employees by how many tasks they have
        ntasks = sorted([
            (
                e.num_tasks + len(assignments[i]),  # Consider idle assignments
                random.random(),  # Random value for tie breaker
                i,
            )
            for i, e in enumerate(self.employees)
        ])

        while len(remaining_tasks) > 0:
            num_tasks, r, employee_id = ntasks[0]
            assignments[employee_id].append(remaining_tasks.pop())
            ntasks[0] = (num_tasks + 1, r, employee_id)

            # Maintain sorted order by swapping up the updated count
            idx = 0
            while idx + 1 < len(ntasks) and ntasks[idx] > ntasks[idx + 1]:
                ntasks[idx + 1], ntasks[idx] = ntasks[idx], ntasks[idx + 1]
                idx += 1

        return assignments

    def schedule_tasks(self, tasks: Sequence[RuntimeTask]) -> None:
        """Schedule tasks between this node's employees."""
        if len(tasks) == 0:
            return

        assignments = self.assign_tasks(tasks)

        for e, assignment in zip(self.employees, assignments):
            num_tasks = len(assignment)

            if num_tasks == 0:
                continue

            self.outgoing.put((e.conn, RuntimeMessage.SUBMIT_BATCH, assignment))

            e.num_tasks += num_tasks
            e.num_idle_workers -= min(num_tasks, e.num_idle_workers)

        self.num_idle_workers = sum(e.num_idle_workers for e in self.employees)

    def send_result_down(self, result: RuntimeResult) -> None:
        """Send the `result` to the appropriate employee."""
        dest_worker_id = result.return_address.worker_id

        if not self.is_my_worker(dest_worker_id):
            raise RuntimeError('Cannot send result to unmanaged worker.')

        employee = self.get_employee_responsible_for(dest_worker_id)
        self.outgoing.put((employee.conn, RuntimeMessage.RESULT, result))

    def is_my_worker(self, worker_id: int) -> bool:
        """Return true if `worker_id` is one of my workers (recursively)."""
        employee_id = (worker_id - self.lower_id_bound) // self.step_size
        return 0 <= employee_id < len(self.employees)

    def get_employee_responsible_for(self, worker_id: int) -> RuntimeEmployee:
        """Return the employee that manages `worker_id`."""
        employee_id = (worker_id - self.lower_id_bound) // self.step_size
        return self.employees[employee_id]

    def broadcast_cancel(self, addr: RuntimeAddress) -> None:
        """Broadcast a cancel message to my employees."""
        for employee in self.employees:
            self.outgoing.put((employee.conn, RuntimeMessage.CANCEL, addr))

    def handle_waiting(self, conn: Connection, new_idle_count: int) -> None:
        """
        Record that an employee is idle with nothing to do.

        There is a race condition here that is allowed. If an employee
        sends a waiting message at the same time that this sends it a
        task, it will still be marked waiting even though it is running
        a task. We allow this for two reasons. First, the consequences are
        minimal: this situation can only lead to one extra task assigned
        to the worker that could otherwise go to a truly idle worker.
        Second, it is unlikely in the common BQSKit workflows, which have
        wide and shallow task graphs and each leaf task can require seconds
        of runtime.
        """
        old_count = self.conn_to_employee_dict[conn].num_idle_workers
        self.conn_to_employee_dict[conn].num_idle_workers = new_idle_count
        self.num_idle_workers += (new_idle_count - old_count)
        assert 0 <= self.num_idle_workers <= self.total_workers


def parse_ipports(ipports_str: Sequence[str]) -> list[tuple[str, int]]:
    """Parse command line ip and port inputs."""
    ipports = []
    for ipport_group in ipports_str:
        for ipport in ipport_group.split(','):
            if ipport.strip() == '':
                continue
            comps = ipport.strip().split(':')

            if len(comps) == 1:
                ip, port = comps[0], str(default_manager_port)
                # Expect only managers to be listening on these ips
                # so default port is manager's default port.

            elif len(comps) == 2:
                ip, port = comps

            else:
                raise ValueError(f'Invalid manager address: {ipport}.')

            if not (0 <= int(port) < 65536):
                raise ValueError(f'Invalid port number: {ipport}')

            ipports.append((ip, int(port)))
    return ipports


def import_tests_package() -> None:
    """
    Import tests package recursively during detached architecture testing.

    This should only be run by the CI test suite from the root bqskit folder.

    credit: https://www.youtube.com/watch?v=t43zBsVcva0
    """
    sys.path.append(os.path.join(os.getcwd()))
    import tests
    import pkgutil
    for mod in pkgutil.walk_packages(tests.__path__, f'{tests.__name__}.'):
        __import__(mod.name, fromlist=['_trash'])



"""This module defines the GreedyPartitioner pass."""
from __future__ import annotations

import bisect
import logging
import warnings

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.interval import CycleInterval
from bqskit.ir.point import CircuitPoint
from bqskit.ir.region import CircuitRegion
from bqskit.utils.typing import is_integer

_logger = logging.getLogger(__name__)


class GreedyPartitioner(BasePass):
    """
    The GreedyPartitioner Pass.

    This pass partitions a circuit by forming the largest regions first.
    """

    def __init__(self, block_size: int = 3) -> None:
        """
        Construct a GreedyPartitioner.

        Args:
            block_size (int): Maximum size of partitioned blocks.
                (Default: 3)

        Raises:
            ValueError: If `block_size` is less than 2.

        Note:
            GreedyPartitioner has been deprecated and will be removed in
            a future update. See :class:`QuickPartitioner` for a
            replacement partitioner.
        """
        warnings.warn(
            'GreedyPartitioner has been deprecated and will be '
            'removed in a future update. See QuickPartitioner '
            'for a replacement partitioner.',
            DeprecationWarning,
        )

        if not is_integer(block_size):
            raise TypeError(
                f'Expected integer for block_size, got {type(block_size)}.',
            )

        if block_size < 2:
            raise ValueError(
                f'Expected block_size to be greater than 2, got {block_size}.',
            )

        self.block_size = block_size

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        if self.block_size > circuit.num_qudits:
            _logger.warning(
                'Configured block size is greater than circuit size; '
                'blocking entire circuit.',
            )
            circuit.fold({
                qudit_index: (0, circuit.num_cycles)
                for qudit_index in range(circuit.num_qudits)
            })
            return

        # For each gate, calculate the best region surrounding it
        total_num_gates = 0
        regions: list[CircuitRegion] = []
        all_bounds: list[list[CycleInterval]] = [
            [] for q in range(circuit.num_qudits)
        ]
        potential_regions = {}
        for cycle, op in circuit.operations_with_cycles():
            point = CircuitPoint(cycle, op.location[0])
            if len(op.location) > self.block_size:
                regions.append(circuit.get_region([point]))
                for qudit, bounds in circuit.get_region([point]).items():
                    bisect.insort(all_bounds[qudit], bounds)
                continue
            total_num_gates += 1
            region = circuit.surround(
                point,
                self.block_size,
                fail_quickly=True,
            )

            potential_regions[point] = (len(circuit[region]), region)

        # Form regions until there are no more gates to partition
        num_partitioned_gates = 0
        while num_partitioned_gates < total_num_gates:

            # Pick largest region
            s = sorted(potential_regions.values(), key=lambda x: x[0])
            num_gates, best_region = s[-1]
            num_partitioned_gates += num_gates
            regions.append(best_region)

            # Update all_bounds
            for qudit, bounds in best_region.items():
                bisect.insort(all_bounds[qudit], bounds)

            # Update others
            to_remove = []
            to_update = []
            for point, value in potential_regions.items():
                region = value[1]

                if point in best_region or region == best_region:
                    to_remove.append(point)
                    continue

                if best_region.overlaps(region):
                    to_update.append(point)
                    continue

            for point in to_remove:
                potential_regions.pop(point)

            for point in to_update:
                # Calculate bounding region
                bounding_region = {}
                cycle = point[0]
                for qudit, bounds_list in enumerate(all_bounds):
                    # find first bound with lower larger than cycle
                    if len(bounds_list) == 0:
                        bounding_region[qudit] = (
                            0, circuit.num_cycles - 1,
                        )
                        continue

                    index_of_first_larger = None
                    for i, bounds in enumerate(bounds_list):
                        if bounds.lower > cycle:
                            index_of_first_larger = i
                            break

                    if index_of_first_larger is None:
                        bounding_region[qudit] = (
                            bounds_list[-1][1] + 1,
                            circuit.num_cycles - 1,
                        )
                    elif index_of_first_larger == 0:
                        bounding_region[qudit] = (0, bounds_list[0][0] - 1)
                    else:
                        bounding_region[qudit] = (
                            bounds_list[index_of_first_larger - 1][1] + 1,
                            bounds_list[index_of_first_larger][0] - 1,
                        )

                    if bounding_region[qudit][0] > bounding_region[qudit][1]:
                        bounding_region.pop(qudit)

                bounding_region = CircuitRegion(bounding_region)

                # Recalculate region
                region = circuit.surround(
                    point,
                    self.block_size,
                    bounding_region,
                    True,
                )

                potential_regions[point] = (len(circuit[region]), region)

        # TODO: Merge regions that can be merged together

        # Fold the circuit
        folded_circuit = Circuit(circuit.num_qudits, circuit.radixes)
        regions = self.topo_sort(regions)
        # Option to keep a block's idle qudits as part of the CircuitGate
        if 'keep_idle_qudits' in data and data['keep_idle_qudits'] is True:
            for region in regions:
                small_region = circuit.downsize_region(region)
                cgc = circuit.get_slice(small_region.points)
                if len(region.location) > len(small_region.location):
                    for i in range(len(region.location)):
                        if region.location[i] not in small_region.location:
                            cgc.insert_qudit(i)
                folded_circuit.append_gate(
                    CircuitGate(cgc, True),
                    sorted(list(region.keys())),
                    list(cgc.params),
                )
        else:
            for region in regions:
                region = circuit.downsize_region(region)
                if 0 < len(region) <= self.block_size:
                    cgc = circuit.get_slice(region.points)
                    folded_circuit.append_gate(
                        CircuitGate(cgc, True),
                        sorted(list(region.keys())),
                        list(cgc.params),
                    )
                else:
                    folded_circuit.extend(circuit[region])
        circuit.become(folded_circuit)

    def topo_sort(self, regions: list[CircuitRegion]) -> list[CircuitRegion]:
        """Topologically sort regions."""
        sorted_regions: list[CircuitRegion] = []
        in_adj_list: list[list[int]] = [[] for _ in range(len(regions))]

        for i, region1 in enumerate(regions):
            for j, region2 in enumerate(regions):
                if i == j:
                    continue
                if region1.depends_on(region2):
                    in_adj_list[i].append(j)  # j points to i

        already_selected: list[int] = []
        while len(already_selected) != len(regions):
            selected = None
            for i, in_nodes in enumerate(in_adj_list):
                if i not in already_selected and len(in_nodes) == 0:
                    selected = i
                    break

            if selected is None:
                raise RuntimeError('Unable to topologically sort regions.')

            sorted_regions.append(regions[selected])
            already_selected.append(selected)
            for in_nodes in in_adj_list:
                if selected in in_nodes:
                    in_nodes.remove(selected)

        return sorted_regions



"""This module implements the Rebase2QuditGatePass (Deprecated Location)."""
from __future__ import annotations

import warnings

warnings.warn(
    'The Rebase2QuditGatePass has moved to bqskit.passes.retarget.two.'
    ' The old location will be removed and this warning will become an'
    ' error in the future.',
    DeprecationWarning,
)

from bqskit.passes.retarget.two import Rebase2QuditGatePass  # noqa



"""This module implements the ChangePredicate class."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from bqskit.passes.control.predicate import PassPredicate

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit


_logger = logging.getLogger(__name__)


class ChangePredicate(PassPredicate):
    """
    The ChangePredicate class.

    The ChangePredicate returns true if the circuit has changed since the last
    call. On the first call, the predicate returns true.
    """

    key = 'ChangePredicate_circuit_hash'

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""

        # If first call, record data and return true
        if self.key not in data:
            _logger.debug(f'Could not find {self.key} key.')
            data[self.key] = self.get_hash(circuit)
            return True

        # Otherwise, check for a change and return accordingly
        new_hash = self.get_hash(circuit)
        if data[self.key] == new_hash:
            # TODO: hash collisions?
            _logger.debug('Hashes match; no change detected.')
            return False

        data[self.key] = new_hash
        _logger.debug('Hashes do not match; change detected.')
        return True

    def get_hash(self, circuit: Circuit) -> int:
        """Retreive hash associated with `circuit`."""
        _logger.debug('Calculating hash for circuit...')

        if len(circuit) == 0:
            return hash(tuple())

        hashes: list[int] = []
        for op in circuit:
            hashes.append(hash(repr(op)))

            # Don't let the hash list grow too large.
            if len(hashes) >= 100:
                hashes = [hash(tuple(hashes))]

        hash_val = hash(tuple(hashes)) if len(hashes) > 1 else hashes[0]
        _logger.debug(f'Hash: {hash_val}')
        return hash_val



"""This module implements the LocallyOptimizableUnitary base class."""
from __future__ import annotations

import abc

import numpy as np
import numpy.typing as npt

from bqskit.qis.unitary.unitary import Unitary
from bqskit.utils.typing import is_square_matrix


class LocallyOptimizableUnitary(Unitary):
    """
    A locally optimizable unitary-valued function.

    A locally optimizable unitary-valued function is one that can be optimized
    with respect to a fixed environment. A `LocallyOptimizableUnitary` inherits
    from `Unitary` and additionally exposes the :func:`optimize` abstract
    method.
    """

    @abc.abstractmethod
    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        More specifically, return the parameters that maximize the
        real component of the trace of the product between `env_matrix`
        and this unitary:

        .. math::

            argmax(Re(Tr(\\mathit{env\\_matrix} \\times
            \\mathit{self.get\\_unitary(params))}))

        Args:
            env_matrix (npt.NDArray[np.complex128]): Optimize with respect
                to this matrix. Has the same dimensions as this unitary.

        Returns:
            list[float]: The parameters that optimizes this unitary.
        """

    def check_env_matrix(self, env_matrix: npt.NDArray[np.complex128]) -> None:
        """Check to ensure the `env_matrix` is validly shaped."""
        if not is_square_matrix(env_matrix):
            raise TypeError('Expected a square matrix.')

        if env_matrix.shape != (self.dim, self.dim):
            raise TypeError('Environmental matrix shape mismatch.')



"""
The Unitary package.

This package exports the Unitary base class and several children
classes: UnitaryMatrix, UnitaryBuilder, DifferentiableUnitary, and
LocallyOptimizableUnitary.
"""
from __future__ import annotations

from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitary import Unitary
from bqskit.qis.unitary.unitarybuilder import UnitaryBuilder
from bqskit.qis.unitary.unitarymatrix import UnitaryLike
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix

__all__ = [
    'Unitary',
    'RealVector',
    'UnitaryLike',
    'UnitaryMatrix',
    'UnitaryBuilder',
    'DifferentiableUnitary',
    'LocallyOptimizableUnitary',
]



"""This module implements the AttachedServer runtime."""
from __future__ import annotations

import logging
import selectors
import uuid
from multiprocessing.connection import Connection
from typing import Any

from bqskit.runtime import default_server_port
from bqskit.runtime import default_worker_port
from bqskit.runtime.base import ServerBase
from bqskit.runtime.detached import DetachedServer
from bqskit.runtime.detached import ServerMailbox
from bqskit.runtime.direction import MessageDirection


class AttachedServer(DetachedServer):
    """
    BQSKit Runtime Server in attached mode.

    In attached mode, the runtime is started by the client. The client owns the
    server and there is only one client on the server. Additionally, the client
    is responsible for shutting down the server. There are no managers in the
    attached architecture, only workers directly managed by a single server.
    This architecture is designed for single-machine shared-memory settings.
    BQSKit will, by default, create, manage, and shutdown an AttachedServer when
    a :class:`~bqskit.compiler.compiler.Compiler` object is created.
    """

    def __init__(
        self,
        num_workers: int = -1,
        port: int = default_server_port,
        worker_port: int = default_worker_port,
    ) -> None:
        """
        Create a server with `num_workers` workers.

        Args:
            num_workers (int): The number of workers to spawn. If -1,
                then spawn as many workers as CPUs on the system.
                (Default: -1).

            port (int): The port this server will listen for clients on.
                Default can be found in the
                :obj:`~bqskit.runtime.default_server_port` global variable.

            worker_port (int): The port this server will listen for workers
                on. Default can be found in the
                :obj:`~bqskit.runtime.default_worker_port` global variable.
        """
        ServerBase.__init__(self)

        # See DetachedServer for more info on the following fields:
        self.clients: dict[Connection, set[uuid.UUID]] = {}
        self.tasks: dict[uuid.UUID, tuple[int, Connection]] = {}
        self.mailbox_to_task_dict: dict[int, uuid.UUID] = {}
        self.mailboxes: dict[int, ServerMailbox] = {}
        self.mailbox_counter = 0

        # Start workers
        self.spawn_workers(num_workers, worker_port)

        # Connect to client
        client_conn = self.listen_once('localhost', port)
        self.clients[client_conn] = set()
        self.sel.register(
            client_conn,
            selectors.EVENT_READ,
            MessageDirection.CLIENT,
        )
        self.logger.info('Connected to client.')

    def handle_disconnect(self, conn: Connection) -> None:
        """A client disconnect in attached mode is equal to a shutdown."""
        self.handle_shutdown()


def start_attached_server(
    num_workers: int,
    log_level: int,
    **kwargs: Any,
) -> None:
    """Start a runtime server in attached mode."""
    # Initialize runtime logging
    _logger = logging.getLogger('bqskit-runtime')
    _logger.setLevel(log_level)
    _logger.addHandler(logging.StreamHandler())

    # Initialize the server
    server = AttachedServer(num_workers, **kwargs)

    # Run the server
    server.run()



"""This module implements the RuntimeResult NamedTuple."""
from __future__ import annotations

from typing import Any
from typing import NamedTuple

from bqskit.runtime.address import RuntimeAddress


class RuntimeResult(NamedTuple):
    """The result of a task, ready to be shipped to its destination."""
    return_address: RuntimeAddress
    result: Any
    completed_by: int



"""This module implements the GeneralizedSabreAlgorithm class."""
from __future__ import annotations

import copy
import logging
from typing import Iterator
from typing import Sequence

import numpy as np

from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.barrier import BarrierPlaceholder
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.gates.constant.swap import SwapGate
from bqskit.ir.operation import Operation
from bqskit.ir.point import CircuitPoint
from bqskit.qis.graph import CouplingGraph


_logger = logging.getLogger(__name__)


class GeneralizedSabreAlgorithm():
    """
    Implements methods for Sabre-based layout and routing algorithms using a
    modified heuristic to accommodate larger than 2-qudit gates.

    References:
        Gushu Li, Yufei Ding, and Yuan Xie. 2019. Tackling the Qubit
        Mapping Problem for NISQ-Era Quantum Devices. In Proceedings of
        the 24th ACM International Conference on Architectural
        Support for Programming Languages and Operating Systems
        (ASPLOS 2019). Association for Computing Machinery, New York, NY,
        USA, 1001-1014. https://doi.org/10.1145/3297858.3304023

        Casey Duckering, Jonathan M. Baker, Andrew Litteken, and Frederic
        T. Chong. 2021. Orchestrated trios: compiling for efficient
        communication in Quantum programs with 3-Qubit gates. In Proceedings
        of the 26th ACM International Conference on Architectural Support
        for Programming Languages and Operating Systems (ASPLOS 2021).
        Association for Computing Machinery, New York, NY, USA, 375-385.
        https://doi.org/10.1145/3445814.3446718
    """

    def __init__(
        self,
        decay_delta: float = 0.001,
        decay_reset_interval: int = 5,
        decay_reset_on_gate: bool = True,
        extended_set_size: int = 20,
        extended_set_weight: float = 0.5,
    ) -> None:
        """
        Construct a GeneralizedSabreAlgorithm.

        Args:
            decay_delta (float): The amount to adjust the decay factor by
                each time a swap is applied. Set to zero to disable decay.
                (Default: 0.001)

            decay_reset_interval (int): The amount of swaps to apply before
                reseting the decay factors. (Default: 5)

            decay_reset_on_gate (bool): If true, reset decay factors when
                a logical gate is applied. (Default: True)

            extended_set_size (int): The size of the look-ahead or extended
                set. Set to zero to disable look ahead. (Default: 20)

            extended_set_weight (float): The weight on the extended set
                term when scoring potential swaps. (Default: 0.5)
        """
        if not isinstance(decay_delta, float):
            raise TypeError(
                'Expected float for decay_delta'
                f', got {type(decay_delta)}',
            )

        if not isinstance(decay_reset_interval, int):
            raise TypeError(
                'Expected int for decay_reset_interval'
                f', got {type(decay_reset_interval)}',
            )

        if not isinstance(decay_reset_on_gate, bool):
            raise TypeError(
                'Expected bool for decay_reset_on_gate'
                f', got {type(decay_reset_on_gate)}',
            )

        if not isinstance(extended_set_size, int):
            raise TypeError(
                'Expected int for extended_set_size'
                f', got {type(extended_set_size)}',
            )

        if not isinstance(extended_set_weight, float):
            raise TypeError(
                'Expected float for extended_set_weight'
                f', got {type(extended_set_weight)}',
            )

        if decay_reset_interval < 1:
            raise ValueError('Decay reset interval must be a positive integer.')

        if extended_set_size < 0:
            raise ValueError('Extended set size must be a nonnegative integer.')

        self.decay_delta = decay_delta
        self.decay_reset_interval = decay_reset_interval
        self.decay_reset_on_gate = decay_reset_on_gate
        self.extended_set_size = extended_set_size
        self.extended_set_weight = extended_set_weight

    def forward_pass(
        self,
        circuit: Circuit,
        pi: list[int],
        cg: CouplingGraph,
        modify_circuit: bool = False,
    ) -> None:
        """
        Apply a forward pass of the Sabre algorithm to `pi`.

        Args:
            circuit (Circuit): The circuit to pass over.

            pi (list[int]): The input logical-to-physical mapping. This
                maps logical qudits to physical qudits. So, `pi[l] == p`
                implies logical qudit `l` is sitting on physical qudit `p`.

            cg (CouplingGraph): The connectivity of the hardware.

            modfiy_circuit (bool): Whether to modify the circuit as the
                pass is applied or not. (Default: False)
        """
        # Preprocessing
        D = cg.all_pairs_shortest_path()
        F = circuit.front
        decay = [1.0 for i in range(circuit.num_qudits)]
        iter_count = 0
        prev_executed_counts: dict[CircuitPoint, int] = {n: 0 for n in F}
        leading_swaps: list[tuple[int, int]] = []
        _logger.debug(f'Starting forward sabre pass with pi: {pi}.')

        if not all(r == circuit.radixes[0] for r in circuit.radixes):
            raise RuntimeError('Cannot currently map to hybrid-level systems.')
        radix = circuit.radixes[0]

        if modify_circuit:
            mapped_circuit = Circuit(circuit.num_qudits, circuit.radixes)

        # Main Loop
        while len(F) > 0:

            # Retrieve executable gates giving the current mapping `pi`
            execute_list = [n for n in F if self._can_exe(circuit[n], pi, cg)]

            # Execute the gates and update F
            if len(execute_list) > 0:
                leading_swaps = []

                for n in execute_list:
                    F.remove(n)
                    prev_executed_counts.pop(n)
                    _logger.debug(f'Executing gate at point {n}.')

                    if modify_circuit:
                        op = circuit[n]
                        physical_location = [pi[q] for q in op.location]
                        mapped_circuit.append_gate(
                            op.gate,
                            physical_location,
                            op.params,
                        )

                    for successor in circuit.next(n):
                        if successor not in prev_executed_counts:
                            prev_executed_counts[successor] = 1
                        else:
                            prev_executed_counts[successor] += 1
                        num_prev_executed = prev_executed_counts[successor]
                        total_num_prev = len(circuit.prev(successor))
                        if num_prev_executed == total_num_prev:
                            F.add(successor)

                # Reset decay if necessary
                if self.decay_reset_on_gate:
                    iter_count = 0
                    for i in range(circuit.num_qudits):
                        decay[i] = 1.0

                continue  # Restart main loop if we executed at least one gate

            # If execute list is empty, check for local-minima
            elif len(leading_swaps) > 5 * cg.num_qudits:
                _logger.debug('Sabre stuck in local minima, backtracking...')

                # Backtrack by removing leading swaps
                for swap in reversed(leading_swaps):
                    self._apply_swap(swap, pi, decay)
                    if modify_circuit:
                        point = mapped_circuit._rear[swap[0]]
                        mapped_circuit.pop(point)
                leading_swaps = []

                # Override heuristic search to progress
                _logger.debug('Overriding sabre search...')
                all_logical_qudits = [circuit[n].location for n in F]
                qudits = min(
                    all_logical_qudits,
                    key=lambda qs: self._get_distance(qs, pi, D),
                )
                for swap in self._uphill_swaps(qudits, cg, pi, D):
                    self._apply_swap(swap, pi, decay)
                    if modify_circuit:
                        mapped_circuit.append_gate(SwapGate(radix), swap)
                _logger.debug('Stopping override.')
                continue

            # Pick and apply a swap
            E = self._calc_extended_set(circuit, F)
            best_swap = self._get_best_swap(circuit, F, E, D, cg, pi, decay)
            self._apply_swap(best_swap, pi, decay)
            leading_swaps.append(best_swap)

            if modify_circuit:
                mapped_circuit.append_gate(SwapGate(radix), best_swap)

            # Update loop counter and reset decay if necessary
            iter_count += 1
            if iter_count % self.decay_reset_interval == 0:
                for i in range(circuit.num_qudits):
                    decay[i] = 1.0

        if modify_circuit:
            circuit.become(mapped_circuit)

    def backward_pass(
        self,
        circuit: Circuit,
        pi: list[int],
        cg: CouplingGraph,
    ) -> None:
        """
        Apply a backward pass of the Sabre algorithm to `pi`.

        Args:
            circuit (Circuit): The circuit to pass over.

            pi (list[int]): The input logical-to-physical mapping. This
                maps logical qudits to physical qudits. So, `pi[l] == p`
                implies logical qudit `l` is sitting on physical qudit `p`.

            cg (CouplingGraph): The connectivity of the hardware.
        """
        # Preprocessing
        D = cg.all_pairs_shortest_path()
        F = circuit.rear
        decay = [1.0 for i in range(circuit.num_qudits)]
        iter_count = 0
        leading_swaps: list[tuple[int, int]] = []
        next_executed_counts: dict[CircuitPoint, int] = {n: 0 for n in F}
        _logger.debug(f'Starting backward sabre pass with pi: {pi}.')

        # Main Loop
        while len(F) > 0:

            # Retrieve executable gates giving the current mapping: pi
            execute_list = [n for n in F if self._can_exe(circuit[n], pi, cg)]

            # Execute the gates and update F
            if len(execute_list) > 0:
                leading_swaps = []

                for n in execute_list:
                    F.remove(n)
                    next_executed_counts.pop(n)
                    _logger.debug(f'Executing gate at point {n}.')

                    for predessor in circuit.prev(n):
                        if predessor not in next_executed_counts:
                            next_executed_counts[predessor] = 1
                        else:
                            next_executed_counts[predessor] += 1
                        num_next_executed = next_executed_counts[predessor]
                        total_num_next = len(circuit.next(predessor))
                        if num_next_executed == total_num_next:
                            F.add(predessor)

                # Reset decay if necessary
                if self.decay_reset_on_gate:
                    iter_count = 0
                    for i in range(circuit.num_qudits):
                        decay[i] = 1.0

                continue  # Restart main loop if we executed at least one gate

            # If execute list is empty, check for local-minima
            elif len(leading_swaps) > 5 * cg.num_qudits:
                _logger.debug('Sabre stuck in local minima, backtracking...')

                # Backtrack by removing leading swaps
                for swap in reversed(leading_swaps):
                    self._apply_swap(swap, pi, decay)
                leading_swaps = []

                # Override heuristic search to progress
                _logger.debug('Overriding sabre search...')
                all_logical_qudits = [circuit[n].location for n in F]
                qudits = min(
                    all_logical_qudits,
                    key=lambda qs: self._get_distance(qs, pi, D),
                )
                for swap in self._uphill_swaps(qudits, cg, pi, D):
                    self._apply_swap(swap, pi, decay)
                _logger.debug('Stopping override.')
                continue

            # Pick and apply a swap
            E = self._calc_extended_set(circuit, F)
            best_swap = self._get_best_swap(circuit, F, E, D, cg, pi, decay)
            self._apply_swap(best_swap, pi, decay)
            leading_swaps.append(best_swap)

            # Update loop counter and reset decay if necessary
            iter_count += 1
            if iter_count % self.decay_reset_interval == 0:
                for i in range(circuit.num_qudits):
                    decay[i] = 1.0

    def _can_exe(self, op: Operation, pi: list[int], cg: CouplingGraph) -> bool:
        """Return true if `op` is executable given the current mapping `pi`."""
        if isinstance(op.gate, BarrierPlaceholder):
            return True

        if isinstance(op.gate, CircuitGate):
            if all(g.num_qudits == 1 for g in op.gate._circuit.gate_set):
                return True

        if op.num_qudits == 1:
            return True

        physical_qudits = [pi[i] for i in op.location]
        return cg.get_subgraph(physical_qudits).is_fully_connected()

    def _calc_extended_set(
        self,
        circuit: Circuit,
        F: set[CircuitPoint],
    ) -> set[CircuitPoint]:
        """Calculate the Extended Set for look-ahead capabilities."""
        extended_set: set[CircuitPoint] = set()
        frontier = list(copy.copy(F))
        while len(frontier) > 0 and len(extended_set) < self.extended_set_size:
            n = frontier.pop(0)
            extended_set.update(circuit.next(n))
            frontier.extend(circuit.next(n))
        return extended_set

    def _get_best_swap(
        self,
        circuit: Circuit,
        F: set[CircuitPoint],
        E: set[CircuitPoint],
        D: list[list[int]],
        cg: CouplingGraph,
        pi: list[int],
        decay: list[float],
    ) -> tuple[int, int]:
        """Return the best swap given the current algorithm state."""
        # Track best one
        best_score = np.inf
        best_swap = None

        # Gather all considerable swaps
        swap_candidate_list = self._obtain_swaps(circuit, F, pi, cg)

        # Score them, tracking the best one
        for swap in swap_candidate_list:
            score = self._score_swap(circuit, F, pi, D, swap, decay, E)
            if score < best_score:
                best_score = score
                best_swap = swap

        if best_swap is None:
            raise RuntimeError('Unable to find best swap.')

        return best_swap

    def _obtain_swaps(
        self,
        circuit: Circuit,
        F: set[CircuitPoint],
        pi: list[int],
        cg: CouplingGraph,
    ) -> set[tuple[int, int]]:
        """Produce all physical swaps with at least one qudit in F."""
        all_qudits: set[int] = set()
        for n in F:
            all_qudits.update(circuit[n].location)
        physical_qudits = [pi[i] for i in all_qudits]

        swaps = set()
        for physical_qudit in physical_qudits:
            neighbors = cg.get_neighbors_of(physical_qudit)
            for neighbor in neighbors:
                a = min(neighbor, physical_qudit)
                b = max(neighbor, physical_qudit)
                swaps.add((a, b))

        return swaps

    def _score_swap(
        self,
        circuit: Circuit,
        F: set[CircuitPoint],
        pi: list[int],
        D: list[list[int]],
        swap: tuple[int, int],
        decay: list[float],
        E: set[CircuitPoint],
    ) -> float:
        """Score the candidate swap given the current algorithm state."""
        # Apply potential swap
        l1, l2 = pi.index(swap[0]), pi.index(swap[1])
        pi[l1], pi[l2] = pi[l2], pi[l1]

        # Calculate front set term
        front = 0.0
        for n in F:
            logical_qudits = circuit[n].location

            # Disallow meaningless swaps
            physical_qudits = [pi[i] for i in logical_qudits]
            if swap[0] in physical_qudits and swap[1] in physical_qudits:
                pi[l1], pi[l2] = pi[l2], pi[l1]
                return np.inf

            front += self._get_distance(logical_qudits, pi, D)
        front /= len(F)

        # Calculate extended set term
        extend = 0.0
        if len(E) > 0:
            for n in E:
                extend += self._get_distance(circuit[n].location, pi, D)
            extend /= len(E)
            extend *= self.extended_set_weight

        # Calculate decay factor
        decay_factor = max(decay[swap[0]], decay[swap[1]])

        # Undo potential swap
        pi[l1], pi[l2] = pi[l2], pi[l1]

        # Return final score
        return decay_factor * (front + extend)

    def _apply_swap(
        self,
        swap: tuple[int, int],
        pi: list[int],
        decay: list[float],
    ) -> None:
        """Apply the swap to `pi` and update `decay`."""
        _logger.debug('applying swap %s' % str(swap))
        l1, l2 = pi.index(swap[0]), pi.index(swap[1])
        pi[l1], pi[l2] = pi[l2], pi[l1]

        decay[swap[0]] += self.decay_delta
        decay[swap[1]] += self.decay_delta

    def _get_distance(
        self,
        logical_qudits: Sequence[int],
        pi: list[int],
        D: list[list[int]],
    ) -> float:
        """Calculate the expected number of swaps to connect logical qudits."""
        min_term = np.inf
        for q in logical_qudits:
            term = 0.0
            for p in logical_qudits:
                if p == q:
                    continue
                term += D[pi[q]][pi[p]]
            min_term = min(term, min_term)
        return min_term

    def _uphill_swaps(
        self,
        logical_qudits: Sequence[int],
        cg: CouplingGraph,
        pi: list[int],
        D: list[list[int]],
    ) -> Iterator[tuple[int, int]]:
        """Yield the swaps necessary to bring some of the qudits together."""
        center_qudit = min(
            logical_qudits,
            key=lambda q: sum(
                D[pi[q]][pi[p]]
                for p in logical_qudits
                if p != q
            ),
        )

        for q in logical_qudits:
            if q == center_qudit:
                continue

            # TODO: Do not need to calculate entire tree
            spt = cg.get_shortest_path_tree(pi[center_qudit])
            path = list(reversed(spt[pi[q]]))

            _logger.debug(f'Moving {q} to {center_qudit} via {path}.')

            for p1, p2 in zip(path, path[1:]):
                if pi[center_qudit] == p1 or pi[center_qudit] == p2:
                    continue
                yield (p1, p2)

    def _apply_perm(self, perm: Sequence[int], pi: list[int]) -> None:
        """Apply the `perm` permutation to the current mapping `pi`."""
        _logger.debug('applying permutation %s' % str(perm))
        pi_c = {q: pi[perm[i]] for i, q in enumerate(sorted(perm))}
        for q in perm:
            pi[q] = pi_c[q]



"""This module implements the HilbertSchmidtCost and
HilbertSchmidtCostGenerator."""
from __future__ import annotations

from typing import TYPE_CHECKING

from bqskitrs import HilbertSchmidtCostFunction

from bqskit.ir.opt.cost.differentiable import DifferentiableCostFunction
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem

if TYPE_CHECKING:
    from bqskit.ir.circuit import Circuit
    from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
    from bqskit.ir.opt.cost.function import CostFunction


class HilbertSchmidtCost(
    HilbertSchmidtCostFunction,
    DifferentiableCostFunction,
):
    """
    The HilbertSchmidtCost CostFunction implementation.

    The Hilbert-Schmidt CostFuction is a differentiable map from circuit
    parameters to a cost value that is based on the Hilbert-Schmidt inner
    product. This function is global-phase-aware, meaning that the cost is zero
    if the target and circuit unitary differ only by a global phase.
    """


class HilbertSchmidtCostGenerator(CostFunctionGenerator):
    """
    The HilbertSchmidtCostGenerator class.

    This generator produces configured HilbertSchmidtCost functions.
    """

    def gen_cost(
        self,
        circuit: Circuit,
        target: UnitaryMatrix | StateVector | StateSystem,
    ) -> CostFunction:
        """Generate a CostFunction, see CostFunctionGenerator for more info."""
        return HilbertSchmidtCost(circuit, target)



"""This module implements the Manager class."""
from __future__ import annotations

import argparse
import logging
import selectors
import time
from multiprocessing.connection import Connection
from typing import Any
from typing import cast
from typing import List
from typing import Sequence

from bqskit.runtime import default_manager_port
from bqskit.runtime import default_worker_port
from bqskit.runtime.address import RuntimeAddress
from bqskit.runtime.base import import_tests_package
from bqskit.runtime.base import parse_ipports
from bqskit.runtime.base import ServerBase
from bqskit.runtime.direction import MessageDirection
from bqskit.runtime.message import RuntimeMessage
from bqskit.runtime.result import RuntimeResult
from bqskit.runtime.task import RuntimeTask


class Manager(ServerBase):
    """
    BQSKit Runtime Manager.

    A Manager is a middle node in the process hierarchy and is responsible for
    managing workers or other managers. The manager is part of the detached
    architecture. Here managers are started individually as separate processes,
    which in turn start their own workers. Then, if necessary, more managers can
    be started to manage the level-1 managers, and so on, until finally, a
    detached server is started and clients can connect.
    """

    def __init__(
        self,
        port: int = default_manager_port,
        num_workers: int = -1,
        ipports: list[tuple[str, int]] | None = None,
        worker_port: int = default_worker_port,
        only_connect: bool = False,
    ) -> None:
        """
        Create a manager instance in one of two ways:

        1) Leave all options default and start a manager which spawns and
            manages as many worker processes as there are os threads.
            You can also specify the number of workers to spawn via
            the `num_workers` parameter. In this mode, the manager
            is a level-1 manager and manages workers.

        2) Specify ip and port pairs, then assume at each endpoint there
            is a listening manager and attempt to establish a connection.
            In this mode, the manager will not spawn any workers and just
            manage the specified managers.

        In either case, if any problems arise during startup, no recovery
        is attempted and the manager terminates.

        Args:
            port (int): The port this manager listens for server connections.
                Default can be found in the
                :obj:`~bqskit.runtime.default_manager_port` global variable.

            num_workers (int): The number of workers to spawn. If -1,
                then spawn as many workers as CPUs on the system.
                (Default: -1). Ignored if `ipports` is not None.

            ipports (list[tuple[str, int]] | None): If not None, then all
                the addresses and ports of running managers to connect to.

            worker_port (int): The port this server will listen for workers
                on. Default can be found in the
                :obj:`~bqskit.runtime.default_worker_port` global variable.

            only_connect (bool): If true, do not spawn workers, only connect
                to already spawned workers.
        """
        super().__init__()

        # Connect upstream
        self.upstream = self.listen_once('0.0.0.0', port)

        # Handshake with upstream
        msg, payload = self.upstream.recv()
        assert msg == RuntimeMessage.CONNECT
        self.lower_id_bound = payload[0]
        self.upper_id_bound = payload[1]
        self.sel.register(
            self.upstream,
            selectors.EVENT_READ,
            MessageDirection.ABOVE,
        )

        # Case 1: spawn and manage workers
        if ipports is None:
            if only_connect:
                self.connect_to_workers(num_workers, worker_port)
            else:
                self.spawn_workers(num_workers, worker_port)

        # Case 2: Connect to managers at ipports
        else:
            self.connect_to_managers(ipports)

        # Track info on sent messages to reduce redundant messages:
        self.last_num_idle_sent_up = self.total_workers

        # Inform upstream we are starting
        msg = (self.upstream, RuntimeMessage.STARTED, self.total_workers)
        self.outgoing.put(msg)
        self.logger.info('Sent start message upstream.')

    def handle_message(
        self,
        msg: RuntimeMessage,
        direction: MessageDirection,
        conn: Connection,
        payload: Any,
    ) -> None:
        """Process the message coming from `direction`."""
        if direction == MessageDirection.ABOVE:

            if msg == RuntimeMessage.SUBMIT:
                rtask = cast(RuntimeTask, payload)
                self.schedule_tasks([rtask])
                self.update_upstream_idle_workers()

            elif msg == RuntimeMessage.SUBMIT_BATCH:
                rtasks = cast(List[RuntimeTask], payload)
                self.schedule_tasks(rtasks)
                self.update_upstream_idle_workers()

            elif msg == RuntimeMessage.RESULT:
                result = cast(RuntimeResult, payload)
                self.send_result_down(result)

            elif msg == RuntimeMessage.CANCEL:
                addr = cast(RuntimeAddress, payload)
                self.broadcast_cancel(addr)

            elif msg == RuntimeMessage.SHUTDOWN:
                self.handle_shutdown()

            else:
                raise RuntimeError(f'Unexpected message type: {msg.name}')

        elif direction == MessageDirection.BELOW:

            if msg == RuntimeMessage.SUBMIT:
                rtask = cast(RuntimeTask, payload)
                self.send_up_or_schedule_tasks([rtask])
                self.update_upstream_idle_workers()

            elif msg == RuntimeMessage.SUBMIT_BATCH:
                rtasks = cast(List[RuntimeTask], payload)
                self.send_up_or_schedule_tasks(rtasks)
                self.update_upstream_idle_workers()

            elif msg == RuntimeMessage.RESULT:
                result = cast(RuntimeResult, payload)
                self.handle_result_from_below(result)

            elif msg == RuntimeMessage.WAITING:
                num_idle = cast(int, payload)
                self.handle_waiting(conn, num_idle)
                self.update_upstream_idle_workers()

            elif msg == RuntimeMessage.UPDATE:
                task_diff = cast(int, payload)
                self.handle_update(conn, task_diff)

            else:
                # Forward all other messages up
                self.outgoing.put((self.upstream, msg, payload))

        else:
            raise RuntimeError(f'Unexpected message from {direction.name}.')

    def handle_system_error(self, error_str: str) -> None:
        """
        Handle an error in runtime code as opposed to client code.

        This is called when an error arises in runtime code not in a
        RuntimeTask's coroutine code.
        """
        try:
            self.upstream.send((RuntimeMessage.ERROR, error_str))

            # Sleep to ensure server receives error message before shutdown
            time.sleep(1)

        except Exception:
            # If server has crashed then just exit
            pass

    def handle_shutdown(self) -> None:
        """Shutdown the manager and clean up spawned processes."""
        super().handle_shutdown()

        # Forward shutdown message upwards if possible
        try:
            self.upstream.send((RuntimeMessage.SHUTDOWN, None))
            self.upstream.close()

        except Exception:
            # If server has already shutdown or crashed, just exit
            pass

    def send_up_or_schedule_tasks(self, tasks: Sequence[RuntimeTask]) -> None:
        """Either send the tasks upstream or schedule them downstream."""
        num_idle = self.num_idle_workers

        if num_idle != 0:
            self.outgoing.put((self.upstream, RuntimeMessage.UPDATE, num_idle))
            self.schedule_tasks(tasks[:num_idle])

        if len(tasks) > num_idle:
            self.outgoing.put((
                self.upstream,
                RuntimeMessage.SUBMIT_BATCH,
                tasks[num_idle:],
            ))

    def handle_result_from_below(self, result: RuntimeResult) -> None:
        """Forward the result to its destination and track the completion."""
        # Record a task has been completed
        self.get_employee_responsible_for(result.completed_by).num_tasks -= 1

        # Forward result to final destination
        if self.is_my_worker(result.return_address.worker_id):
            self.send_result_down(result)
            self.outgoing.put((self.upstream, RuntimeMessage.UPDATE, -1))

        else:
            # If its destination worker is not an employee of mine,
            # then my boss will know where to send this result.
            self.outgoing.put((self.upstream, RuntimeMessage.RESULT, result))

    def update_upstream_idle_workers(self) -> None:
        """Update the total number of idle workers upstream."""
        if self.num_idle_workers != self.last_num_idle_sent_up:
            self.last_num_idle_sent_up = self.num_idle_workers
            m = (self.upstream, RuntimeMessage.WAITING, self.num_idle_workers)
            self.outgoing.put(m)

    def handle_update(self, conn: Connection, task_diff: int) -> None:
        """Handle a task count update from a lower level manager or worker."""
        self.conn_to_employee_dict[conn].num_tasks += task_diff
        self.outgoing.put((self.upstream, RuntimeMessage.UPDATE, task_diff))


def start_manager() -> None:
    """Entry point for runtime manager processes."""
    parser = argparse.ArgumentParser(
        prog='bqskit-manager',
        description='Launch a BQSKit runtime manager process.',
    )
    parser.add_argument(
        '-n', '--num-workers',
        default=-1,
        type=int,
        help='The number of workers to spawn. If negative, will spawn'
        ' one worker for each available CPU. Defaults to -1.',
    )
    parser.add_argument(
        '-m', '--managers',
        nargs='+',
        help='The ip and port pairs were managers are expected to be waiting.',
    )
    parser.add_argument(
        '-p', '--port',
        type=int,
        default=default_manager_port,
        help='The port this manager will listen for servers on.',
    )
    parser.add_argument(
        '-w', '--worker-port',
        type=int,
        default=default_worker_port,
        help='The port this manager will listen for workers on.',
    )
    parser.add_argument(
        '-x', '--only-connect',
        action='store_true',
        help='Do not spawn workers, only connect to them.',
    )
    parser.add_argument(
        '-v', '--verbose',
        action='count',
        default=0,
        help='Enable logging of increasing verbosity, either -v, -vv, or -vvv.',
    )
    parser.add_argument(
        '-i', '--import-tests',
        action='store_true',
        help='Import the bqskit tests package; used during testing.',
    )
    args = parser.parse_args()

    # If ips and ports were provided parse them
    ipports = None if args.managers is None else parse_ipports(args.managers)

    # Set up logging
    _logger = logging.getLogger('bqskit-runtime')
    _logger.setLevel([30, 20, 10, 1][min(args.verbose, 3)])
    _logger.addHandler(logging.StreamHandler())

    # Import tests package recursively
    if args.import_tests:
        import_tests_package()

    if args.num_workers < -1:
        num_workers = -1
    else:
        num_workers = args.num_workers

    # Create the manager
    manager = Manager(
        args.port,
        num_workers,
        ipports,
        args.worker_port,
        args.only_connect,
    )

    # Start the manager
    manager.run()



"""This module implements the CircuitPoint class."""
from __future__ import annotations

import logging
from typing import Any
from typing import Tuple
from typing import Union

from typing_extensions import TypeGuard

from bqskit.utils.typing import is_integer
_logger = logging.getLogger(__name__)


class CircuitPoint(Tuple[int, int]):
    """
    A cycle and qudit index pair used to index a circuit.

    This is a subclass of a tuple, and therefore can be used where a tuple of
    two ints can be used.
    """

    def __new__(
        cls,
        cycle_or_tuple: int | tuple[int, int],
        qudit: int | None = None,
    ) -> CircuitPoint:
        """
        Construct a point.

        Args:
            cycle_or_tuple (int | tuple[int, int]): Either a cycle index
                given as an integer, or a tuple of a cycle index and qudit
                index. If an integer is given, then you will need to also
                specify the qudit index as the next argument.

            qudit (int | None): If `cycle_or_tuple` is an integer cycle
                index, then you will need to specify the qudit index here.
                Otherwise, leave this as None.

        Returns:
            CircuitPoint: The new point object.
        """

        if qudit is not None and not is_integer(qudit):
            raise TypeError(
                f'Expected int or None for qudit, got {type(qudit)}.',
            )

        if isinstance(cycle_or_tuple, tuple):
            if not CircuitPoint.is_point(cycle_or_tuple):
                raise TypeError('Expected two integer arguments.')

            if qudit is not None:
                raise ValueError('Unable to handle extra argument.')

            cycle = cycle_or_tuple[0]
            qudit = cycle_or_tuple[1]

        elif is_integer(cycle_or_tuple) and is_integer(qudit):
            cycle = cycle_or_tuple

        elif is_integer(cycle_or_tuple) and qudit is None:
            raise ValueError('Expected two integer arguments.')

        else:
            raise TypeError('Expected two integer arguments.')

        return super().__new__(cls, (cycle, qudit))  # type: ignore

    @property
    def cycle(self) -> int:
        """The point's cycle index."""
        return self[0]

    @property
    def qudit(self) -> int:
        """The point's qudit index."""
        return self[1]

    @staticmethod
    def is_point(point: Any) -> TypeGuard[CircuitPointLike]:
        """Return true if point is a CircuitPointLike."""
        if isinstance(point, CircuitPoint):
            return True

        if not isinstance(point, tuple):
            _logger.log(0, 'Point is not a tuple.')
            return False

        if len(point) != 2:
            _logger.log(
                0,
                'Expected point to contain two values, got %d.' % len(
                    point,
                ),
            )
            return False

        if not is_integer(point[0]):
            _logger.log(
                0,
                'Expected integer values in point, got %s.' % type(
                    point[0],
                ),
            )
            return False

        if not is_integer(point[1]):
            _logger.log(
                0,
                'Expected integer values in point, got %s.' % type(
                    point[1],
                ),
            )
            return False

        return True


CircuitPointLike = Union[Tuple[int, int], CircuitPoint]



"""This module implements the MultiPhysicalPredicate class."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from bqskit.passes.control.predicate import PassPredicate

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit

_logger = logging.getLogger(__name__)


class MultiPhysicalPredicate(PassPredicate):
    """
    The MultiPhysicalPredicate class.

    The MultiPhysicalPredicate returns true if circuit's multi-qudit gates are
    in the native gate set.
    """

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        model = data.model
        for gate in circuit.gate_set:
            if gate.num_qudits < 2:
                continue
            if gate not in model.gate_set:
                _logger.debug(f'{gate} not in native set: {model.gate_set}.')
                return False

        return True



"""This module implements the ZXZXZDecomposition."""
from __future__ import annotations

import cmath

import numpy as np

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.constant.sx import SqrtXGate
from bqskit.ir.gates.parameterized.rx import RXGate
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.ir.gates.parameterized.u1 import U1Gate


class ZXZXZDecomposition(BasePass):
    """
    The ZXZXZDecomposition class.

    Convert a single-qubit circuit to ZXZXZ sequence.
    """

    def __init__(
        self,
        always_use_rx: bool = False,
        always_use_u1: bool = False,
    ) -> None:
        """
        Construct a ZXZXZDecomposition pass.

        Args:
            always_use_rx (bool): If True, always use RX instead of SX.

            always_use_u1 (bool): If True, always use U1 instead of RZ.
        """

        if not isinstance(always_use_rx, bool):
            raise TypeError(
                f'Expected bool for always_use_rx, got {type(always_use_rx)}.',
            )

        if not isinstance(always_use_u1, bool):
            raise TypeError(
                f'Expected bool for always_use_u1, got {type(always_use_u1)}.',
            )

        self.always_use_rx = always_use_rx
        self.always_use_u1 = always_use_u1

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        if circuit.num_qudits != 1:
            raise ValueError(
                'Cannot convert multi-qudit circuit into ZXZXZ sequence.',
            )

        if circuit.radixes[0] != 2:
            raise ValueError(
                'Cannot convert non-qubit circuit into ZXZXZ sequence.',
            )

        # Decide on RX or SX
        no_sx = RXGate() in data.gate_set and SqrtXGate() not in data.gate_set
        use_rx = self.always_use_rx or no_sx

        # Decide on RZ or U1
        no_rz = U1Gate() in data.gate_set and RZGate() not in data.gate_set
        use_u1 = self.always_use_u1 or no_rz

        utry = circuit.get_unitary()

        # Calculate params
        utry = np.linalg.det(utry) ** (-0.5) * utry
        i1 = cmath.phase(utry[1, 1])
        i2 = cmath.phase(utry[1, 0])
        t = 2 * np.arctan2(abs(utry[1, 0]), abs(utry[0, 0])) + np.pi
        p = i1 + i2 + np.pi
        l = i1 - i2

        # Move angles into [-pi, pi)
        t = (t + np.pi) % (2 * np.pi) - np.pi
        p = (p + np.pi) % (2 * np.pi) - np.pi
        l = (l + np.pi) % (2 * np.pi) - np.pi

        new_circuit = Circuit(1)

        if use_u1:
            new_circuit.append_gate(U1Gate(), 0, [l])
        else:
            new_circuit.append_gate(RZGate(), 0, [l])

        if use_rx:
            new_circuit.append_gate(RXGate(), 0, [np.pi / 2])
        else:
            new_circuit.append_gate(SqrtXGate(), 0)

        if use_u1:
            new_circuit.append_gate(U1Gate(), 0, [t])
        else:
            new_circuit.append_gate(RZGate(), 0, [t])

        if use_rx:
            new_circuit.append_gate(RXGate(), 0, [np.pi / 2])
        else:
            new_circuit.append_gate(SqrtXGate(), 0)

        if use_u1:
            new_circuit.append_gate(U1Gate(), 0, [p])
        else:
            new_circuit.append_gate(RZGate(), 0, [p])

        circuit.become(new_circuit)



"""This module implements the ComposedGate class."""
from __future__ import annotations

from bqskit.ir.gate import Gate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary


class ComposedGate(Gate):
    """
    A gate composed of other gates.

    A ComposedGate provides methods for determining if the gate is
    differentiable or locally optimizable.

    A ComposedGate is differentiable/locally optimizable if
    it inherits from the appropriate base class and all of its subgates
    (in either self.gate or self.gates) inherit from the base class.

    For more complex behaviors, one can override :func:`is_differentiable`
    or :func:`is_locally_optimizable`.
    """

    def is_differentiable(self) -> bool:
        """Check if all sub gates are differentiable."""
        if hasattr(self, 'gate'):
            return isinstance(self.gate, DifferentiableUnitary)
        if hasattr(self, 'gates'):
            return all(
                isinstance(gate, DifferentiableUnitary)
                for gate in self.gates
            )

        raise AttributeError(
            'Expected gate or gates field for composed gate %s.'
            % self.name,
        )

    def is_locally_optimizable(self) -> bool:
        """Check if all sub gates are locally optimizable."""
        if hasattr(self, 'gate'):
            return isinstance(
                self.gate, LocallyOptimizableUnitary,
            )
        if hasattr(self, 'gates'):
            return all(
                isinstance(gate, LocallyOptimizableUnitary)
                for gate in self.gates
            )

        raise AttributeError(
            'Expected gate or gates field for composed gate %s.'
            % self.name,
        )

    def __hash__(self) -> int:
        if hasattr(self, 'gate'):
            return hash((self.name, self.gate))
        if hasattr(self, 'gates'):
            return hash((self.name, tuple(self.gates)))
        raise RuntimeError(
            f"Composed gate '{self.name}' has no attribute 'gate' or 'gates'.",
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self.__dict__ == other.__dict__



"""This module implements the CeresMinimizer class."""
from __future__ import annotations

import logging
from typing import TypeVar

from bqskitrs import LeastSquaresMinimizerNative

from bqskit.ir.opt.minimizer import Minimizer

_logger = logging.getLogger(__name__)

# TODO: use PEP 673 Self type when mypy is upgraded
Self = TypeVar('Self', bound='CeresMinimizer')


class CeresMinimizer(LeastSquaresMinimizerNative, Minimizer):
    """
    The CeresMinimizer class.

    The CeresMinimizer attempts to instantiate the circuit such that the
    circuit's cost, given by a CostFunction, is minimized.
    """

    def __new__(
        cls: type[Self], num_threads: int = 1, ftol: float = 1e-6,
        gtol: float = 1e-10, report: bool = False,
    ) -> Self:
        """
        Create a new CeresMinimizer.

        Args:
            num_threads (int): The number of threads Ceres should use
                for optimization (defaults to 1)

            ftol (float): Default 1e-6. The function tolerance stopping
                condition Ceres should use, see the documentation about
                ftol here:
                http://ceres-solver.org/nnls_solving.html#_CPPv4N5ceres6Solver7Options18function_toleranceE

            gtol (float): Default 1e-10. The gradient tolerance stopping
                condition Ceres should use, see the documentation about
                gtol here:
                http://ceres-solver.org/nnls_solving.html#_CPPv4N5ceres6Solver7Options18gradient_toleranceE

            report (bool): Whether to print the minimization summary Ceres
                can generate. This is quite noisy so only enable this if
                you are certain you want it! Defaults to false.
        """

        return super().__new__(cls, num_threads, ftol, gtol, report)



"""This module implements the CircuitLocation class."""
from __future__ import annotations

import logging
from typing import Any
from typing import Iterator
from typing import overload
from typing import Sequence
from typing import Union

from typing_extensions import TypeGuard

from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_iterable
from bqskit.utils.typing import is_mapping
from bqskit.utils.typing import is_sequence
from bqskit.utils.typing import is_sequence_of_int
_logger = logging.getLogger(__name__)


class CircuitLocation(Sequence[int]):
    """
    The CircuitLocation class.

    A CircuitLocation is an ordered set of qudit indices. In context, this
    usually describes where a gate or operation is being applied.
    """

    def __init__(self, location: int | Sequence[int]) -> None:
        """
        Construct a CircuitLocation from `location`.

        Args:
            location (int | Sequence[int]): The qudit indices.

        Raises:
            ValueError: If any qudit index is negative.

            ValueError: If there are duplicates in location.
        """

        if is_integer(location):
            location = [location]

        elif is_sequence(location):
            location = list(location)

        else:
            raise TypeError(
                f'Expected integer(s) for location, got {type(location)}.',
            )

        if not is_sequence_of_int(location):
            checks = [is_integer(q) for q in location]
            raise TypeError(
                'Expected iterable of positive integers for location'
                f', got atleast one {type(location[checks.index(False)])}.',
            )

        if not all(qudit_index >= 0 for qudit_index in location):
            checks = [qudit_index >= 0 for qudit_index in location]
            raise ValueError(
                'Expected iterable of positive integers for location'
                f', got atleast one {location[checks.index(False)]}.',
            )

        if len(set(location)) != len(location):
            raise ValueError('Location has duplicate qudit indices.')

        self._location: tuple[int, ...] = tuple(location)

    @overload
    def __getitem__(self, index: int) -> int:
        ...

    @overload
    def __getitem__(self, indices: slice | Sequence[int]) -> tuple[int, ...]:
        ...

    def __getitem__(
        self,
        index: int | slice | Sequence[int],
    ) -> int | tuple[int, ...]:
        """Retrieve one or multiple qudit indices from the location."""
        if is_integer(index):
            return self._location[index]

        if isinstance(index, slice):
            return self._location[index]

        if is_sequence_of_int(index):
            return tuple(self._location[idx] for idx in index)

        raise TypeError(f'Invalid index type, got {type(index)}.')

    def __len__(self) -> int:
        """Return the number of qudits described by the location."""
        return len(self._location)

    def union(self, other: CircuitLocationLike) -> CircuitLocation:
        """Return the location containing qudits from self or other."""
        if is_integer(other):
            if other in self:
                return self
            return CircuitLocation(self._location + (other,))
        return CircuitLocation(list(set(self).union(CircuitLocation(other))))

    def intersection(self, other: CircuitLocationLike) -> CircuitLocation:
        """Return the location containing qudits from self and other."""
        if is_integer(other):
            return CircuitLocation([other] if other in self else [])
        return CircuitLocation([x for x in self if x in other])  # type: ignore

    @property
    def pairs(self) -> set[tuple[int, int]]:
        """Return all pairs of unique elements in the location."""
        _pairs = set()
        for q1 in self:
            for q2 in self:
                if q1 == q2:
                    continue
                a = min(q1, q2)
                b = max(q1, q2)
                _pairs.add((a, b))
        return _pairs

    @staticmethod
    def is_location(
        location: Any,
        num_qudits: int | None = None,
    ) -> TypeGuard[CircuitLocationLike]:
        """
        Determines if the sequence of qudits form a valid location. A valid
        location is a set of qubit indices (integers) that are greater than or
        equal to zero, and if num_qudits is specified, less than num_qudits.

        Args:
            location (Any): The location to check.

            num_qudits (int | None): The total number of qudits.
                All qudit indices should be less than this. If None,
                don't check.

        Returns:
            (TypeGuard[CircuitLocationLike]): True if the location is valid.
        """
        if isinstance(location, CircuitLocation):
            if num_qudits is not None:
                return max(location) < num_qudits
            return True

        if is_mapping(location) or isinstance(location, set):
            return False

        if is_integer(location):
            location = [location]

        elif is_iterable(location):
            location = list(location)

        else:
            _logger.log(
                0,
                f'Expected integer(s) for location, got {type(location)}.',
            )
            return False

        if not is_sequence_of_int(location):
            checks = [is_integer(q) for q in location]
            _logger.log(
                0,
                'Expected iterable of positive integers for location'
                f', got atleast one {type(location[checks.index(False)])}.',
            )
            return False

        if not all(qudit_index >= 0 for qudit_index in location):
            checks = [qudit_index >= 0 for qudit_index in location]
            _logger.log(
                0,
                'Expected iterable of positive integers for location'
                f', got atleast one {location[checks.index(False)]}.',
            )
            return False

        if len(set(location)) != len(location):
            _logger.log(0, 'Location has duplicate qudit indices.')
            return False

        if num_qudits is not None:
            if not all([qudit < num_qudits for qudit in location]):
                _logger.log(0, 'Location has an erroneously large qudit.')
                return False

        return True

    def __str__(self) -> str:
        return self._location.__str__()

    def __repr__(self) -> str:
        return self._location.__repr__()

    def __hash__(self) -> int:
        return self._location.__hash__()

    def __iter__(self) -> Iterator[int]:
        return self._location.__iter__()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CircuitLocation):
            return self._location.__eq__(other._location)

        if is_integer(other):
            return len(self) == 1 and self[0] == other

        return self._location.__eq__(other)


CircuitLocationLike = Union[int, Sequence[int], CircuitLocation]



"""This package implements BQSKit's post-processing passes."""
from __future__ import annotations

from bqskit.passes.processing.exhaustive import ExhaustiveGateRemovalPass
from bqskit.passes.processing.iterative import IterativeScanningGateRemovalPass
from bqskit.passes.processing.rebase import Rebase2QuditGatePass  # TODO: Remove
from bqskit.passes.processing.scan import ScanningGateRemovalPass
from bqskit.passes.processing.substitute import SubstitutePass

__all__ = [
    'ExhaustiveGateRemovalPass',
    'IterativeScanningGateRemovalPass',
    'ScanningGateRemovalPass',
    'SubstitutePass',
    'Rebase2QuditGatePass',
]



"""This module implements the various checkpointing passes."""
from __future__ import annotations

import logging
import pickle

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit

_logger = logging.getLogger(__name__)


class SaveCheckpointPass(BasePass):
    """
    The SaveCheckpointPass class.

    The SaveCheckpointPass saves the current circuit as a data file.
    """

    def __init__(self, checkpoint_filename: str) -> None:
        """
        Constructor for the SaveCheckpointPass.

        Args:
            checkpoint_filename (str): Full path name for the checkpoint.
        """
        self.checkpoint_filename = checkpoint_filename

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see BasePass for more info."""
        with open(self.checkpoint_filename, 'wb') as f:
            pickle.dump((circuit, data), f)


class LoadCheckpointPass(BasePass):
    """
    The LoadCheckpointPass class.

    The LoadCheckpointPass loads the circuit from a data file.
    """

    def __init__(self, checkpoint_filename: str) -> None:
        """
        Constructor for the SaveCheckpointPass.

        Args:
            checkpoint_filename (str): Full path name for the checkpoint.
        """
        self.checkpoint_filename = checkpoint_filename

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see BasePass for more info."""
        with open(self.checkpoint_filename, 'rb') as f:
            checkpoint = pickle.load(f)
            circuit.become(checkpoint[0])
            data.clear()
            data.update(checkpoint[1])



"""This module implements BQSKit Runtime's Worker."""
from __future__ import annotations

import argparse
import logging
import os
import signal
import sys
import time
import traceback
from collections import OrderedDict
from dataclasses import dataclass
from multiprocessing import Process
from multiprocessing.connection import Client
from multiprocessing.connection import Connection
from multiprocessing.connection import wait
from typing import Any
from typing import Callable
from typing import cast
from typing import List

from bqskit.runtime import default_worker_port
from bqskit.runtime.address import RuntimeAddress
from bqskit.runtime.future import RuntimeFuture
from bqskit.runtime.message import RuntimeMessage
from bqskit.runtime.result import RuntimeResult
from bqskit.runtime.task import RuntimeTask


class WorkerQueue():
    """The worker's task FIFO queue."""

    def __init__(self) -> None:
        """
        Initialize the worker queue.

        An OrderedDict is used to internally store the task. This prevents the
        same task appearing multiple times in the queue, while also ensuring
        O(1) operations.
        """
        self._queue: OrderedDict[RuntimeAddress, None] = OrderedDict()

    def put(self, addr: RuntimeAddress) -> None:
        """Enqueue a task by its address."""
        if addr not in self._queue:
            self._queue[addr] = None

    def get(self) -> RuntimeAddress:
        """Get the next task to run."""
        return self._queue.popitem(last=False)[0]

    def empty(self) -> bool:
        """Check if the queue is empty."""
        return len(self._queue) == 0


@dataclass
class WorkerMailbox:
    """
    A mailbox on a worker is a final destination for a task's result.

    When a task is created, a mailbox is also created with an associated future.
    The parent task can await on the future, letting the worker's event loop
    know it is waiting on the associated result. When a result arrives, it is
    placed in the appropriate mailbox and the waiting task is placed into the
    ready queue.
    """
    expecting_single_result: bool = False
    expected_num_results: int = 0
    result: Any = None
    num_results: int = 0
    dest_addr: RuntimeAddress | None = None
    fresh_results: list[Any] | None = None

    @property
    def ready(self) -> bool:
        """Return true if the mailbox has all expected results."""
        return (
            self.num_results >= self.expected_num_results
            and self.num_results != 0
        )

    @property
    def has_task_waiting(self) -> bool:
        """Return True if a task is waiting on the result of this box."""
        return self.dest_addr is not None

    @staticmethod
    def new_mailbox(num_results: int | None = None) -> WorkerMailbox:
        """
        Create a new mailbox with `num_results` slots.

        If `num_results` is None (by default), then the mailbox will only have
        one slot and expect one result.
        """
        if num_results is None:
            return WorkerMailbox(True, 1)

        return WorkerMailbox(False, num_results, [None] * num_results)

    def get_new_results(self) -> list[tuple[int, Any]]:
        """Return and reset the results that have come in since previous
        call."""
        assert self.fresh_results is not None
        out = self.fresh_results
        self.fresh_results = []
        return out

    def deposit_result(self, result: RuntimeResult) -> None:
        """Store the result in the mailbox."""
        self.num_results += 1
        slot_id = result.return_address.mailbox_slot

        # Record as fresh result
        if self.fresh_results is None:
            self.fresh_results = []
        self.fresh_results.append((slot_id, result.result))

        if self.expecting_single_result:
            self.result = result.result
        else:
            self.result[slot_id] = result.result


class Worker:
    """
    BQSKit Runtime's Worker.

    BQSKit Runtime utilizes a single-threaded worker to accept, execute,
    pause, spawn, resume, and complete tasks in a custom event loop built
    with python's async await mechanisms. Each worker receives and sends
    tasks and results to the greater system through a single duplex
    connection with a runtime server or manager.

    At start-up, the worker receives an ID and waits for its first task.
    An executing task may use the `submit` and `map` methods to spawn child
    tasks and distribute them across the whole system. Once completed,
    those child tasks will have their results shipped back to the worker
    who created them. When a task awaits a child task, it is removed from
    the ready queue until the desired results come in.

    All created log records are shipped back to the client's process.
    This feature ensures compatibility with applications like jupyter
    that only print messages from the client process's stdout. Additionally,
    it allows BQSKit users seamless integration with the standard python
    logging module. From a user's perspective, they can configure any
    standard python logger from their process like usual and have the
    entire system honor that configuration. Lastly, we do support an
    additional logging option for maximum task depth. Tasks with more
    ancestors than the maximum logging depth will not produce any logs.

    Workers handle python errors by capturing and bubbling them up. For
    operating system-level crashes and errors -- such as seg-faults in
    client code -- the worker will attempt to print a stack trace and
    initiate a system-wide shutdown. However, note that these issues
    can go unhandled and cause the runtime to deadlock or crash.

    Workers perform very minimal scheduling of tasks. Newly created tasks
    are directly forwarded upwards to a manager or server, which in turn
    assigns them to workers. New tasks from above are commonly received
    in batches. In this case, all but one task from a batch is delayed.
    Delayed tasks are moved (in LIFO order) to the ready queue if a worker
    has no other work to complete. This mechanism encourages completing
    deeply-nested tasks first and prevents flooding the system with active
    tasks, which usually require much more memory than delayed ones.
    """

    def __init__(self, id: int, conn: Connection) -> None:
        """
        Initialize a worker with no tasks.

        Args:
            id (int): This worker's id.

            conn (Connection): This worker's duplex channel to a manager
                or a server.
        """
        self._id = id
        self._conn = conn

        self._outgoing: list[tuple[RuntimeMessage, Any]] = []
        """Stores outgoing messages to be handled by the event loop."""

        self._tasks: dict[RuntimeAddress, RuntimeTask] = {}
        """Tracks all started, unfinished tasks on this worker."""

        self._delayed_tasks: list[RuntimeTask] = []
        """Store all delayed tasks in LIFO order."""

        self._ready_task_ids: WorkerQueue = WorkerQueue()
        """Tasks queued up for execution."""

        self._cancelled_task_ids: set[RuntimeAddress] = set()
        """To ensure newly-received cancelled tasks are never started."""

        self._active_task: RuntimeTask | None = None
        """The currently executing task if one is running."""

        self._running = False
        """Controls if the event loop is running."""

        self._mailboxes: dict[int, WorkerMailbox] = {}
        """Map from mailbox ids to worker mailboxes."""

        self._mailbox_counter = 0
        """This count ensures every mailbox has a unique id."""

        # Send out every emitted log message upstream
        old_factory = logging.getLogRecordFactory()

        def record_factory(*args: Any, **kwargs: Any) -> logging.LogRecord:
            record = old_factory(*args, **kwargs)
            active_task = get_worker()._active_task
            if active_task is not None:
                lvl = active_task.logging_level
                if lvl is None or lvl <= record.levelno:
                    tid = active_task.comp_task_id
                    self._outgoing.append((RuntimeMessage.LOG, (tid, record)))
            return record

        logging.setLogRecordFactory(record_factory)

        # Communicate that this worker is ready
        self._conn.send((RuntimeMessage.STARTED, self._id))

    def _loop(self) -> None:
        """Main worker event loop."""
        self._running = True
        while self._running:
            self._try_step_next_ready_task()
            self._try_idle()
            self._handle_comms()

    def _try_idle(self) -> None:
        """If there is nothing to do, wait until we receive a message."""
        empty_outgoing = len(self._outgoing) == 0
        no_ready_tasks = self._ready_task_ids.empty()
        no_delayed_tasks = len(self._delayed_tasks) == 0

        if empty_outgoing and no_ready_tasks and no_delayed_tasks:
            self._conn.send((RuntimeMessage.WAITING, 1))
            wait([self._conn])

    def _handle_comms(self) -> None:
        """Handle all incoming and outgoing messages."""

        # Handle outgoing communication
        for out_msg in self._outgoing:
            self._conn.send(out_msg)
        self._outgoing.clear()

        # Handle incomming communication
        while self._conn.poll():
            msg, payload = self._conn.recv()

            # Process message
            if msg == RuntimeMessage.SHUTDOWN:
                self._running = False
                return

            elif msg == RuntimeMessage.SUBMIT:
                task = cast(RuntimeTask, payload)
                self._add_task(task)

            elif msg == RuntimeMessage.SUBMIT_BATCH:
                tasks = cast(List[RuntimeTask], payload)
                self._add_task(tasks.pop())  # Submit one task
                self._delayed_tasks.extend(tasks)  # Delay rest
                # Delayed tasks have no context and are stored (more-or-less)
                # as a function pointer together with the arguments.
                # When it gets started, it consumes much more memory,
                # so we delay the task start until necessary (at no cost)

            elif msg == RuntimeMessage.RESULT:
                result = cast(RuntimeResult, payload)
                self._handle_result(result)

            elif msg == RuntimeMessage.CANCEL:
                addr = cast(RuntimeAddress, payload)
                self._handle_cancel(addr)

    def _add_task(self, task: RuntimeTask) -> None:
        """Start a task and add it to the loop."""
        self._tasks[task.return_address] = task
        task.start()
        self._ready_task_ids.put(task.return_address)

    def _handle_result(self, result: RuntimeResult) -> None:
        """Insert result into appropriate mailbox and wake waiting task."""
        mailbox_id = result.return_address.mailbox_index
        assert result.return_address.worker_id == self._id
        if mailbox_id not in self._mailboxes:
            # If the mailbox has been dropped due to a cancel, ignore result
            return

        box = self._mailboxes[mailbox_id]
        box.deposit_result(result)

        if box.has_task_waiting:
            assert box.dest_addr is not None
            task = self._tasks[box.dest_addr]

            if task.wake_on_next or box.ready:
                self._ready_task_ids.put(box.dest_addr)  # Wake it

    def _handle_cancel(self, addr: RuntimeAddress) -> None:
        """
        Remove `addr` and its children tasks from this worker.

        Notes:
            Since `self._ready_task_ids' is a queue, it is more efficient
            to discard cancelled tasks when popping from it. Therefore, we
            do not do anything with `self._ready_task_ids` here.

            Also, we also don't need to send out cancel messages for
            cancelled children tasks since other workers can evaluate that
            for themselves using breadcrumbs and the original `addr` cancel
            message.
        """
        self._cancelled_task_ids.add(addr)

        # Remove all tasks that are children of `addr` from initialized tasks
        for key, task in self._tasks.items():
            if task.is_descendant_of(addr):
                for mailbox_id in self._tasks[key].owned_mailboxes:
                    self._mailboxes.pop(mailbox_id)
        self._tasks = {
            a: t for a, t in self._tasks.items()
            if not t.is_descendant_of(addr)
        }

        # Remove all tasks that are children of `addr` from delayed tasks
        self._delayed_tasks = [
            t for t in self._delayed_tasks
            if not t.is_descendant_of(addr)
        ]

    def _get_next_ready_task(self) -> RuntimeTask | None:
        """Return the next ready task if one exists, otherwise None."""
        while True:
            if self._ready_task_ids.empty():
                if len(self._delayed_tasks) > 0:
                    self._add_task(self._delayed_tasks.pop())
                    continue
                return None

            addr = self._ready_task_ids.get()

            if addr in self._cancelled_task_ids or addr not in self._tasks:
                # When a task is cancelled on the worker it is not removed
                # from the ready queue because it is much cheaper to just
                # discard cancelled tasks as they come out.
                continue

            task = self._tasks[addr]

            if any(bcb in self._cancelled_task_ids for bcb in task.breadcrumbs):
                # If any of the selected tasks ancestor tasks are cancelled
                # then discard this one too. Each breadcrumb (bcb) is a
                # task address (unique system-wide task id) of an ancestor
                # task.
                continue

            return task

    def _try_step_next_ready_task(self) -> None:
        """Select a task to run, and advance it one step."""
        task = self._get_next_ready_task()

        if task is None:
            # Nothing to do
            return

        try:
            self._active_task = task

            # Perform a step of the task and get the future it awaits on
            future = task.step(self._get_desired_result(task))

            self._process_await(task, future)

        except StopIteration as e:
            self._process_task_completion(task, e.value)

        except Exception:
            assert self._active_task is not None  # for type checker

            # Bubble up errors
            exc_info = sys.exc_info()
            error_str = ''.join(traceback.format_exception(*exc_info))
            error_payload = (self._active_task.comp_task_id, error_str)
            self._outgoing.append((RuntimeMessage.ERROR, error_payload))

        finally:
            self._active_task = None

    def _process_await(self, task: RuntimeTask, future: RuntimeFuture) -> None:
        """Process a task's await request."""
        if not isinstance(future, RuntimeFuture):
            raise RuntimeError('Can only await on a BQSKit RuntimeFuture.')

        if future.mailbox_id not in self._mailboxes:
            raise RuntimeError('Cannot await on a canceled task.')

        box = self._mailboxes[future.mailbox_id]

        # Let the mailbox know this task is waiting
        box.dest_addr = task.return_address
        task.desired_box_id = future.mailbox_id

        if future._next_flag:
            # Set from Worker.next, implies the task wants the next result
            if box.ready:
                m = 'Cannot wait for next results on a complete task.'
                raise RuntimeError(m)
            task.wake_on_next = True

        elif box.ready:
            self._ready_task_ids.put(task.return_address)

    def _process_task_completion(self, task: RuntimeTask, result: Any) -> None:
        """Package and send out task result."""
        assert task is self._active_task
        packaged_result = RuntimeResult(task.return_address, result, self._id)

        if task.return_address.worker_id == self._id:
            self._handle_result(packaged_result)
            self._outgoing.append((RuntimeMessage.UPDATE, -1))
            # Let manager know this worker has one less task
            # without sending a result
        else:
            self._outgoing.append((RuntimeMessage.RESULT, packaged_result))

        # Remove task
        self._tasks.pop(task.return_address)

        # Cancel any open tasks
        for mailbox_id in self._active_task.owned_mailboxes:
            # If task is complete, simply discard result
            if mailbox_id in self._mailboxes:
                if self._mailboxes[mailbox_id].ready:
                    self._mailboxes.pop(mailbox_id)
                    continue

            # Otherwise send a cancel message
            self.cancel(RuntimeFuture(mailbox_id))

        # Start delayed task
        if self._ready_task_ids.empty() and len(self._delayed_tasks) > 0:
            self._add_task(self._delayed_tasks.pop())

    def _get_desired_result(self, task: RuntimeTask) -> Any:
        """Retrieve the task's desired result from the mailboxes."""
        if task.desired_box_id is None:
            return None

        box = self._mailboxes[task.desired_box_id]

        if task.wake_on_next:
            fresh_results = box.get_new_results()
            assert len(fresh_results) > 0
            return fresh_results

        assert box.ready
        task.owned_mailboxes.remove(task.desired_box_id)
        return self._mailboxes.pop(task.desired_box_id).result

    def _get_new_mailbox_id(self) -> int:
        """Return a new unique mailbox id."""
        new_id = self._mailbox_counter
        self._mailbox_counter += 1
        return new_id

    def submit(
        self,
        fn: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> RuntimeFuture:
        """Submit `fn` as a task to the runtime."""
        assert self._active_task is not None
        # Group fnargs together
        fnarg = (fn, args, kwargs)

        # Create a new mailbox
        mailbox_id = self._get_new_mailbox_id()
        self._mailboxes[mailbox_id] = WorkerMailbox.new_mailbox()
        self._active_task.owned_mailboxes.append(mailbox_id)

        # Create the task
        task = RuntimeTask(
            fnarg,
            RuntimeAddress(self._id, mailbox_id, 0),
            self._active_task.comp_task_id,
            self._active_task.breadcrumbs + (self._active_task.return_address,),
            self._active_task.logging_level,
            self._active_task.max_logging_depth,
        )

        # Submit the task (on the next cycle)
        self._outgoing.append((RuntimeMessage.SUBMIT, task))

        # Return future pointing to the mailbox
        return RuntimeFuture(mailbox_id)

    def map(
        self,
        fn: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> RuntimeFuture:
        """Map `fn` over the input arguments distributed across the runtime."""
        assert self._active_task is not None
        # Group fnargs together
        fnargs = []
        if len(args) == 1:
            for arg in args[0]:
                fnargs.append((fn, (arg,), kwargs))

        else:
            for subargs in zip(*args):
                fnargs.append((fn, subargs, kwargs))

        if len(fnargs) == 0:
            raise RuntimeError('Unable to map 0 tasks.')

        # Create a new mailbox
        mailbox_id = self._get_new_mailbox_id()
        self._mailboxes[mailbox_id] = WorkerMailbox.new_mailbox(len(fnargs))
        self._active_task.owned_mailboxes.append(mailbox_id)

        # Create the tasks
        breadcrumbs = self._active_task.breadcrumbs
        breadcrumbs += (self._active_task.return_address,)
        tasks = [
            RuntimeTask(
                fnarg,
                RuntimeAddress(self._id, mailbox_id, i),
                self._active_task.comp_task_id,
                breadcrumbs,
                self._active_task.logging_level,
                self._active_task.max_logging_depth,
            )
            for i, fnarg in enumerate(fnargs)
        ]

        # Submit the tasks
        self._outgoing.append((RuntimeMessage.SUBMIT_BATCH, tasks))

        # Return future pointing to the mailbox
        return RuntimeFuture(mailbox_id)

    def cancel(self, future: RuntimeFuture) -> None:
        """Cancel all tasks associated with `future`."""
        assert self._active_task is not None
        num_slots = self._mailboxes[future.mailbox_id].expected_num_results
        self._active_task.owned_mailboxes.remove(future.mailbox_id)
        self._mailboxes.pop(future.mailbox_id)
        addrs = [
            RuntimeAddress(self._id, future.mailbox_id, slot_id)
            for slot_id in range(num_slots)
        ]
        msgs = [(RuntimeMessage.CANCEL, addr) for addr in addrs]
        self._outgoing.extend(msgs)

    async def next(self, future: RuntimeFuture) -> list[tuple[int, Any]]:
        """
        Wait for and return the next batch of results from a map task.

        Returns:
            (list[tuple[int, Any]]): A list of the results that arrived
                since the last time this was called. On the first call,
                all results that have arrived since the task started are
                returned. Each result is paired with the index of its
                arguments in the original map call.
        """
        if future._done:
            raise RuntimeError('Cannot wait on an already completed result.')

        future._next_flag = True
        next_result_batch = await future
        future._next_flag = False
        return next_result_batch


# Global variable containing reference to this process's worker object.
_worker = None


def start_worker(w_id: int | None, port: int, cpu: int | None = None) -> None:
    """Start this process's worker."""
    if w_id is not None:
        # Ignore interrupt signals on workers, boss will handle it for us
        # If w_id is None, then we are being spawned separately.
        signal.signal(signal.SIGINT, signal.SIG_IGN)

    # Purge all standard python logging configurations
    for _, logger in logging.Logger.manager.loggerDict.items():
        if isinstance(logger, logging.PlaceHolder):
            continue
        logger.handlers.clear()
    logging.Logger.manager.loggerDict = {}

    if cpu is not None:
        if sys.platform == 'win32':
            raise RuntimeError('Cannot pin worker to cpu on windows.')
        os.sched_setaffinity(0, [cpu])

    max_retries = 7
    wait_time = .1
    conn: Connection | None = None
    family = 'AF_INET' if sys.platform == 'win32' else None
    for _ in range(max_retries):
        try:
            conn = Client(('localhost', port), family)
        except (ConnectionRefusedError, TimeoutError):
            time.sleep(wait_time)
            wait_time *= 2
        else:
            break

    if conn is None:
        raise RuntimeError('Unable to establish connection with manager.')

    if w_id is None:
        msg, w_id = conn.recv()
        assert msg == RuntimeMessage.STARTED

    global _worker
    _worker = Worker(w_id, conn)
    _worker._loop()


def get_worker() -> Worker:
    """Return a handle on this process's worker."""
    if _worker is None:
        raise RuntimeError('Worker has not been started.')
    return _worker


def _check_positive(value: str) -> int:
    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError(
            '%s is an invalid positive int value' % value,
        )
    return ivalue


def start_worker_rank() -> None:
    """Entry point for spawning a rank of runtime worker processes."""
    parser = argparse.ArgumentParser(
        prog='bqskit-worker',
        description='Launch a rank of BQSKit runtime worker processes.',
    )
    parser.add_argument(
        'num_workers',
        type=_check_positive,
        help='The number of workers to spawn.',
    )
    parser.add_argument(
        '--cpus', '-c',
        nargs='+',
        type=int,
        help='Either one number or a list of numbers equal in length to the'
        ' number of workers. The workers will be pinned to specified logical'
        ' cpus. If a single-number is given, then all cpu indices are'
        ' enumerated starting at that number.',
    )
    parser.add_argument(
        '-p', '--port',
        type=int,
        default=default_worker_port,
        help='The port the workers will try to connect to a manager on.',
    )
    args = parser.parse_args()

    if args.cpus is not None:
        if len(args.cpus) == 1:
            cpus = [args.cpus[0] + i for i in range(args.num_workers)]

        elif len(args.cpus) == args.num_workers:
            cpus = args.cpus

        else:
            raise RuntimeError(
                'The specified logical cpus are invalid. Expected either'
                ' a single number or a list of numbers equal in length to'
                ' the number of workers.',
            )

    else:
        cpus = [None for _ in range(args.num_workers)]

    # Spawn worker process
    procs = []
    for cpu in cpus:
        procs.append(Process(target=start_worker, args=(None, args.port, cpu)))
        procs[-1].start()

    # Join them
    for proc in procs:
        proc.join()



"""This module implements the PermutationMatrix class."""
from __future__ import annotations

import logging
from typing import Sequence

import numpy as np

from bqskit.qis.unitary.unitarybuilder import UnitaryBuilder
from bqskit.qis.unitary.unitarymatrix import UnitaryLike
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer
_logger = logging.getLogger(__name__)


class PermutationMatrix(UnitaryMatrix):
    """A binary, unitary matrix with a single 1 in each row and column."""

    def __init__(
        self,
        input: UnitaryLike,
        radixes: Sequence[int] = [],
        check_arguments: bool = True,
    ) -> None:
        """
        Constructs a `PermutationMatrix` from the supplied matrix.

        See :class:`UnitaryMatrix` for more info.
        """
        if check_arguments and not PermutationMatrix.is_permutation(input):
            raise ValueError('Input failed permutation condition.')

        super().__init__(input, radixes, check_arguments)

    @staticmethod
    def from_qubit_location(
        num_qubits: int,
        location: Sequence[int],
    ) -> PermutationMatrix:
        """
        Creates a qubit permutation matrix.

        See :func:`from_qudit_location` for more info.
        """
        return PermutationMatrix.from_qudit_location(num_qubits, 2, location)

    @staticmethod
    def from_qudit_location(
        num_qudits: int,
        radix: int,
        location: Sequence[int],
    ) -> PermutationMatrix:
        """
        Creates the permutation matrix specified by arguments.

        The resulting matrix will move the location's qudits into positions
        defined by their index in location. For example:

        P = Permutationmatrix.from_qudit_location(3, 2, (1, 2, 0))

                .-----.
             0 -|     |- 1
             1 -|  P  |- 2
             2 -|     |- 0
                '-----'

        Args:
            num_qudits (int): Total number of qudits.

            radix (int): The base of all qudits.

            location (Sequence[int]): The desired qudits to swap
                to their index.

        Returns:
            (PermutationMatrix): A permutation matrix that moves the
                location's qudits into positions defined by their index
                in location.
        """
        swap_utry = PermutationMatrix.gen_swap_unitary(radix)
        current_perm = list(location)
        for i in range(num_qudits):
            if i not in current_perm:
                current_perm.append(i)

        perm_builder = UnitaryBuilder(num_qudits, [radix] * num_qudits)
        for index, qudit in enumerate(current_perm):
            if index != qudit:
                current_pos = current_perm.index(index)
                perm_builder.apply_left(swap_utry, (index, current_pos))
                tmp = current_perm[index]
                current_perm[index] = current_perm[current_pos]
                current_perm[current_pos] = tmp
        return PermutationMatrix(perm_builder.get_unitary())

    @staticmethod
    def is_permutation(P: np.typing.ArrayLike, tol: float = 1e-8) -> bool:
        """
        Check if P is a permutation matrix.

        Args:
            P (np.typing.ArrayLike): The matrix to check.

            tol (float): The numerical precision of the check.

        Returns:
            bool: True if P is a permutation matrix.
        """

        if isinstance(P, PermutationMatrix):
            return True

        if not UnitaryMatrix.is_unitary(P, tol):
            return False

        if not isinstance(P, np.ndarray):
            P = np.array(P)

        if not all(s == 1 for s in P.sum(0)):
            _logger.debug('Not all rows sum to 1.')
            return False

        if not all(s == 1 for s in P.sum(1)):
            _logger.debug('Not all columns sum to 1.')
            return False

        if not all(e == 1 or e == 0 for row in P for e in row):
            _logger.debug('Not all elements are 0 or 1.')
            return False

        return True

    @staticmethod
    def gen_swap_unitary(radix: int = 2) -> UnitaryMatrix:
        """
        Generate a unitary matrix that swaps the state of two qudits.

        Args:
            radix (int): The base of the qudits being swapped.
                Defaults to qubits or base 2. (Default: 2)

        Raises:
            ValueError: If radix is less than two.
        """
        if not is_integer(radix):
            raise TypeError('Expected a single integer radix.')

        if radix < 2:
            raise ValueError('Radix must be at least 2.')

        dim = radix * radix
        mat = [[0 for _j in range(dim)] for _i in range(dim)]
        for col in range(dim):
            # col = a * radix + b; a, b < radix
            a = col // radix
            b = col % radix
            row = b * radix + a
            mat[row][col] = 1

        return UnitaryMatrix(mat, [radix, radix])



"""This module implements the YYGate."""
from __future__ import annotations

import math

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class YYGate(ConstantGate, QubitGate):
    """
    The Ising YY coupling gate.

    The YY gate is given by the following unitary:

    .. math::

        \\begin{pmatrix}
        \\frac{\\sqrt{2}}{2} & 0 & 0 & \\frac{\\sqrt{2}}{2}i \\\\
        0 & \\frac{\\sqrt{2}}{2} & -\\frac{\\sqrt{2}}{2}i & 0 \\\\
        0 & -\\frac{\\sqrt{2}}{2}i & \\frac{\\sqrt{2}}{2} & 0 \\\\
        \\frac{\\sqrt{2}}{2}i & 0 & 0 & \\frac{\\sqrt{2}}{2} \\\\
        \\end{pmatrix}
    """

    _num_qudits = 2
    _qasm_name = 'ryy(pi/2)'
    _utry = UnitaryMatrix(
        [
            [math.sqrt(2) / 2, 0, 0, 1j * math.sqrt(2) / 2],
            [0, math.sqrt(2) / 2, -1j * math.sqrt(2) / 2, 0],
            [0, -1j * math.sqrt(2) / 2, math.sqrt(2) / 2, 0],
            [1j * math.sqrt(2) / 2, 0, 0, math.sqrt(2) / 2],
        ],
    )



"""This module implements the FrozenParameterGate."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.ir.gates.composedgate import ComposedGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_real_number


class FrozenParameterGate(
    ComposedGate,
    LocallyOptimizableUnitary,
    DifferentiableUnitary,
):
    """A composed gate which fixes some parameters of another gate."""

    def __init__(self, gate: Gate, frozen_params: dict[int, float]) -> None:
        """
        Create a gate which fixes some of the parameters it takes.

        Args:
            gate (Gate): The Gate to fix the parameters of.

            frozen_params (dict[int, float]): A dictionary mapping parameters
                indices to the fixed value they should be.

        Raises:
            ValueError: If any of the `frozen_params` indices are greater
                than the number of parameters `gate` takes or less than 0
                or if the total amount of `frozen_params` is larger than
                the number of parameters `gate` takes.
        """
        if not isinstance(gate, Gate):
            raise TypeError('Expected gate, got %s.' % type(gate))
        if not isinstance(frozen_params, dict):
            raise TypeError(
                'Expected dict for frozen_params, '
                'got %s.' % type(frozen_params),
            )
        if not len(frozen_params) <= gate.num_params:
            raise ValueError(
                'Too many fixed parameters specified, expected at most'
                ' %d, got %d' % (gate.num_params, len(frozen_params)),
            )
        keys = list(frozen_params.keys())
        values = list(frozen_params.values())
        if not all(is_integer(p) for p in keys):
            fail_idx = [is_integer(p) for p in keys].index(False)
            raise TypeError(
                'Expected frozen_params keys to be int, got %s.'
                % type(keys[fail_idx]),
            )
        if not all(is_real_number(p) for p in values):
            typechecks = [is_real_number(p) for p in values]
            fail_idx = typechecks.index(False)
            raise TypeError(
                'Expected frozen_params values to be float, got %s.'
                % type(values[fail_idx]),
            )
        if not all(0 <= p < gate.num_params for p in keys):
            fail_idx = [
                0 <= p < gate.num_params
                for p in keys
            ].index(False)
            raise ValueError(
                'Expected parameter index to be non-negative integer'
                ' < %d, got %d.' % (gate.num_params, keys[fail_idx]),
            )

        self.gate = gate
        self._num_params = gate.num_params - len(frozen_params)
        self._num_qudits = gate.num_qudits
        self._radixes = gate.radixes
        self.frozen_params = frozen_params
        self.unfixed_param_idxs = [
            i for i in range(gate.num_params)
            if i not in self.frozen_params.keys()
        ]
        self._name = '{}({}, {})'.format(
            self.__class__.__name__,
            self.gate.name,
            str(self.frozen_params),
        )
        # TODO: If new gate is constant; cache its unitary

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        return self.gate.get_unitary(self.get_full_params(params))

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        grads = self.gate.get_grad(self.get_full_params(params))  # type: ignore
        return grads[self.unfixed_param_idxs, :, :]

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        f_params = self.get_full_params(params)

        utry, grads = self.gate.get_unitary_and_grad(f_params)  # type: ignore
        return utry, grads[self.unfixed_param_idxs, :, :]

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        params = self.gate.optimize(env_matrix)  # type: ignore
        return [
            p for i, p in enumerate(params)
            if i in self.unfixed_param_idxs
        ]

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, FrozenParameterGate)
            and self.gate == other.gate
            and self.frozen_params == other.frozen_params
        )

    def __hash__(self) -> int:
        return hash((self.gate, tuple(self.frozen_params.items())))

    @property
    def qasm_name(self) -> str:
        """The qasm command for this gate, see Gate for more info."""
        return self.gate.qasm_name

    def get_qasm_gate_def(self) -> str:
        """Returns a qasm def for this gate, see :class:Gate for more."""
        return self.gate.get_qasm_gate_def()

    def get_full_params(self, params: RealVector) -> list[float]:
        """
        Returns the full parameter list for the underlying gate.

        Args:
            params (RealVector): The parameters to the gate.

        Returns:
            list[float]: The full parameters to the underlying gate.
        """
        self.check_parameters(params)
        args = list(params)
        for idx in sorted(self.frozen_params):
            args.insert(idx, self.frozen_params[idx])
        return args


def with_frozen_params(
        self: Gate,
        frozen_params: dict[int, float],
) -> FrozenParameterGate:
    """
    Freeze some of a gate's parameters so they don't change from optimization.

    Args:
        frozen_params (dict[int, float]): A map from parameter indices to
            parameters values. If i in frozen_params, then this will freeze
            the i-th parameter to the value given by frozen_params[i].

    Returns:
        FrozenParameterGate: The gate with some parameters frozen.
    """
    return FrozenParameterGate(self, frozen_params)


def with_all_frozen_params(
    self: Gate,
    params: list[float],
) -> FrozenParameterGate:
    """
    Freeze all of a gate's parameters so they don't change from optimization.

    Args:
        params (list[float]): The values to set and freeze all parameters to.

    Returns:
        FrozenParameterGate: The gate with all parameters frozen.
    """
    return FrozenParameterGate(self, {i: x for i, x in enumerate(params)})


Gate.with_frozen_params = with_frozen_params
Gate.with_all_frozen_params = with_all_frozen_params



"""This package implements synthesis passes and synthesis related classes."""
from __future__ import annotations

from bqskit.passes.synthesis.leap import LEAPSynthesisPass
from bqskit.passes.synthesis.pas import PermutationAwareSynthesisPass
from bqskit.passes.synthesis.qfast import QFASTDecompositionPass
from bqskit.passes.synthesis.qpredict import QPredictDecompositionPass
from bqskit.passes.synthesis.qsearch import QSearchSynthesisPass
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.passes.synthesis.target import SetTargetPass

__all__ = [
    'LEAPSynthesisPass',
    'QFASTDecompositionPass',
    'QPredictDecompositionPass',
    'QSearchSynthesisPass',
    'SynthesisPass',
    'SetTargetPass',
    'PermutationAwareSynthesisPass',
]



"""This module implements the UnitaryMatrix class."""
from __future__ import annotations

import logging
from typing import Any
from typing import Sequence
from typing import TYPE_CHECKING
from typing import Union

import numpy as np
import numpy.typing as npt
import scipy as sp
from scipy.stats import unitary_group

from bqskit.qis.state.state import StateLike
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.statemap import StateVectorMap
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitary import Unitary
from bqskit.utils.docs import building_docs
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_square_matrix
from bqskit.utils.typing import is_valid_radixes

if TYPE_CHECKING:
    from typing import TypeGuard

if not building_docs():
    from numpy.lib.mixins import NDArrayOperatorsMixin
else:
    class NDArrayOperatorsMixin:  # type: ignore
        pass

_logger = logging.getLogger(__name__)


class UnitaryMatrix(Unitary, StateVectorMap, NDArrayOperatorsMixin):
    """A concrete representation of a unitary matrix."""

    def __init__(
        self,
        input: UnitaryLike,
        radixes: Sequence[int] = [],
        check_arguments: bool = True,
    ) -> None:
        """
        Constructs a `UnitaryMatrix` from the supplied unitary matrix.

        Args:
            input (UnitaryLike): The unitary matrix input.

            radixes (Sequence[int]): A sequence with its length equal to
                the number of qudits this `UnitaryMatrix` can act on. Each
                element specifies the base, number of orthogonal states,
                for the corresponding qudit. By default, the constructor
                will attempt to calculate `radixes` from `utry`.

            check_arguments (bool): If true, check arguments for type
                and value errors.

        Raises:
            ValueError: If `input` is not unitary.

            ValueError: If the dimension of `input` does not match the
                expected dimension from `radixes`.

            RuntimeError: If `radixes` is not specified and the
                constructor cannot infer it.

        Examples:
            >>> UnitaryMatrix(
            ...     [
            ...         [0, 1],
            ...         [1, 0],
            ...     ],
            ... )  # Creates a single-qubit Pauli X unitary matrix.
            array([[0.+0.j, 1.+0.j],
                   [1.+0.j, 0.+0.j]])
        """

        # Stop any actual logic when building documentation
        if building_docs():
            self._utry: npt.NDArray[np.complex128] = np.array([])
            return

        # Copy constructor
        if isinstance(input, UnitaryMatrix):
            self._utry = input.numpy
            self._radixes = input.radixes
            self._dim = input.dim
            return

        if check_arguments and not is_square_matrix(input):
            raise TypeError(f'Expected square matrix, got {type(input)}.')

        if check_arguments and not UnitaryMatrix.is_unitary(input):
            raise ValueError('Input failed unitary condition.')

        dim = len(input)

        if radixes:
            self._radixes = tuple(radixes)

        # Check if unitary dimension is a power of two
        elif dim & (dim - 1) == 0:
            self._radixes = tuple([2] * int(np.round(np.log2(dim))))

        # Check if unitary dimension is a power of three
        elif 3 ** int(np.round(np.log(dim) / np.log(3))) == dim:  # noqa
            radixes = [3] * int(np.round(np.log(dim) / np.log(3)))
            self._radixes = tuple(radixes)

        else:
            raise RuntimeError(
                'Unable to determine radixes'
                ' for UnitaryMatrix with dim %d.' % dim,
            )

        if check_arguments and not is_valid_radixes(self.radixes):
            raise TypeError('Invalid qudit radixes.')

        if check_arguments and np.prod(self.radixes) != dim:
            raise ValueError('Qudit radixes mismatch with dimension.')

        self._utry = np.array(input, dtype=np.complex128)
        self._dim = dim

    _num_params = 0

    @property
    def numpy(self) -> npt.NDArray[np.complex128]:
        """The NumPy array holding the unitary."""
        return self._utry

    @property
    def shape(self) -> tuple[int, ...]:
        """The two-dimensional square shape of the unitary."""
        return self._utry.shape

    @property
    def dtype(self) -> np.typing.DTypeLike:
        """The NumPy data type of the unitary."""
        return self._utry.dtype

    @property
    def T(self) -> UnitaryMatrix:
        """The transpose of the unitary."""
        return UnitaryMatrix(self._utry.T, self.radixes, False)

    @property
    def dagger(self) -> UnitaryMatrix:
        """The conjugate transpose of the unitary."""
        return self.conj().T

    def to_special(self) -> UnitaryMatrix:
        """Return a special unitary matrix verson of this one."""
        determinant = np.linalg.det(self)
        dimension = len(self)
        global_phase = np.angle(determinant) / dimension
        global_phase = global_phase % (2 * np.pi / dimension)
        global_phase_factor = np.exp(-1j * global_phase)
        return global_phase_factor * self

    def is_special(self) -> bool:
        """Return true if this unitary is special."""
        return 1 - np.abs(np.linalg.det(self)) < 1e-8

    def __len__(self) -> int:
        """The dimension of the square unitary matrix."""
        return self.shape[0]

    def __iter__(self) -> int:
        """An iterator that iterates through the rows of the matrix."""
        return self._utry.__iter__()

    def conj(self) -> UnitaryMatrix:
        """Return the complex conjugate unitary matrix."""
        return UnitaryMatrix(self._utry.conj(), self.radixes, False)

    def otimes(self, *utrys: UnitaryLike) -> UnitaryMatrix:
        """
        Calculate the tensor or kroneckor product with other unitaries.

        Args:
            utrys (UnitaryLike | Sequence[UnitaryLike]): The unitary or
                unitaries to computer the tensor product with.

        Returns:
            UnitaryMatrix: The resulting unitary matrix.
        """

        utrys = [UnitaryMatrix(u) for u in utrys]

        utry_acm = self.numpy
        radixes_acm = self.radixes
        for utry in utrys:
            utry_acm = np.kron(utry_acm, utry.numpy)
            radixes_acm += utry.radixes

        return UnitaryMatrix(utry_acm, radixes_acm)

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the same object, satisfies the :class:`Unitary` API."""
        return self

    def get_distance_from(self, other: UnitaryLike, degree: int = 2) -> float:
        """
        Return the distance between `self` and `other`.

        The distance is given as:

        .. math::

            \\sqrt[D]{1 - \\frac{|Tr(U_1^\\dagger U_2)|}{N}^D}

        where D is the degree, by default is 2.

        Args:
            other (UnitaryLike): The unitary to measure distance from.

            degree (int): The degree of the distance metric.

        Returns:
            float: A value between 1 and 0, where 0 means the two unitaries
            are equal up to global phase and 1 means the two unitaries are
            very unsimilar or far apart.
        """
        other = UnitaryMatrix(other, check_arguments=False)
        num = np.abs(np.trace(self.conj().T @ other))
        dem = self.dim
        frac = min(num / dem, 1)
        dist = np.power(1 - (frac ** degree), 1.0 / degree)
        return dist if dist > 0.0 else 0.0

    def get_statevector(self, in_state: StateLike) -> StateVector:
        """
        Calculate the output state after applying this unitary to `in_state`.

        Args:
            in_state (StateLike): The input state to apply `self` to.

        Returns:
            StateVector: The output state.

        Raises:
            ValueError: If the state's dimension doesn't match the
                unitary's dimension.
        """
        if not StateVector.is_pure_state(in_state):
            raise TypeError(f'Expected StateVector, got {type(in_state)}.')

        in_state = StateVector(in_state)

        if in_state.dim != self.dim:
            raise ValueError(
                'State unitary dimension mismatch; '
                f'expected {self.dim}, got {in_state.dim}.',
            )

        vec = in_state[:, None]
        out_vec = self @ vec
        return StateVector(out_vec.reshape((-1,)), self.radixes)

    @staticmethod
    def identity(dim: int, radixes: Sequence[int] = []) -> UnitaryMatrix:
        """
        Construct an identity UnitaryMatrix.

        Args:
            dim (int): The dimension of the identity matrix.

            radixes (Sequence[int]): The number of orthogonal states
                for each qudit, defaults to qubits.

        Returns:
            UnitaryMatrix: An identity matrix.

        Raises:
            ValueError: If `dim` is nonpositive.
        """
        if dim <= 0:
            raise ValueError('Invalid dimension for identity matrix.')
        return UnitaryMatrix(np.identity(dim), radixes)

    @staticmethod
    def closest_to(
        M: npt.NDArray[np.complex128],
        radixes: Sequence[int] = [],
    ) -> UnitaryMatrix:
        """
        Calculate and return the closest unitary to a given matrix.

        Calculate the unitary matrix U that is closest with respect to the
        operator norm distance to the general matrix M.

        Args:
            M (np.ndarray): The matrix input.

            radixes (Sequence[int]): The radixes for the Unitary.

        Returns:
            (UnitaryMatrix): The unitary matrix closest to M.

        References:
            D.M.Reich. “Characterisation and Identification of Unitary Dynamics
            Maps in Terms of Their Action on Density Matrices”
        """
        if not is_square_matrix(M):
            raise TypeError('Expected square matrix.')

        V, _, Wh = sp.linalg.svd(M)
        return UnitaryMatrix(V @ Wh, radixes, False)

    @staticmethod
    def random(num_qudits: int, radixes: Sequence[int] = []) -> UnitaryMatrix:
        """
        Sample a random unitary from the haar distribution.

        Args:
            num_qudits (np.ndarray): The number of qudits for the matrix.
                This is not the dimension.

            radixes (Sequence[int]): The radixes for the Unitary.

        Returns:
            (UnitaryMatrix): A random unitary matrix.

        Raises:
            ValueError: If `num_qudits` is nonpositive.

            ValueError: If the length of `radixes` is not equal to
                `num_qudits`.
        """
        if not is_integer(num_qudits):
            raise TypeError(
                f'Expected int for num_qudits, got {type(num_qudits)}.',
            )

        if num_qudits <= 0:
            raise ValueError('Expected positive number for num_qudits.')

        radixes = tuple(radixes if len(radixes) > 0 else [2] * num_qudits)

        if not is_valid_radixes(radixes):
            raise TypeError('Invalid qudit radixes.')

        if len(radixes) != num_qudits:
            raise ValueError(
                'Expected length of radixes to be equal to num_qudits:'
                ' %d != %d' % (len(radixes), num_qudits),
            )

        U = unitary_group.rvs(int(np.prod(radixes)))
        return UnitaryMatrix(U, radixes, False)

    def __eq__(self, other: object) -> bool:
        """Check if `self` is approximately equal to `other`."""
        if isinstance(other, Unitary):
            other_unitary = other.get_unitary()
            if self.shape != other_unitary.shape:
                return False
            return np.allclose(self, other_unitary)

        if isinstance(other, np.ndarray):
            return np.allclose(self, other)

        return NotImplemented

    def save(self, filename: str) -> None:
        """Save the unitary to a file."""
        np.savetxt(filename, self.numpy)

    def __getitem__(
            self, index: Any,
    ) -> np.complex128 | npt.NDArray[np.complex128]:
        """Implements NumPy API for the StateVector class."""
        return self._utry[index]

    @staticmethod
    def from_file(filename: str) -> UnitaryMatrix:
        """Load a unitary from a file."""
        return UnitaryMatrix(np.loadtxt(filename, dtype=np.complex128))

    @staticmethod
    def is_unitary(U: Any, tol: float = 1e-8) -> TypeGuard[UnitaryLike]:
        """
        Check if U is a unitary matrix.

        Args:
            U (Any): The matrix to check.

            tol (float): The numerical precision of the check.

        Returns:
            bool: True if U is a unitary.
        """

        if isinstance(U, UnitaryMatrix):
            return True

        from bqskit.qis.state import StateSystem
        if isinstance(U, StateSystem):
            return False

        if not isinstance(U, np.ndarray):
            U = np.array(U)

        if not is_square_matrix(U):
            return False

        X = U @ U.conj().T
        Y = U.conj().T @ U
        I = np.identity(X.shape[0])

        if not np.allclose(X, I, rtol=0, atol=tol):
            if _logger.isEnabledFor(logging.DEBUG):
                norm = np.linalg.norm(X - I)
                _logger.debug(
                    'Failed unitary condition, ||UU^d - I|| = %e' %
                    norm,
                )
            return False

        if not np.allclose(Y, I, rtol=0, atol=tol):
            if _logger.isEnabledFor(logging.DEBUG):
                norm = np.linalg.norm(Y - I)
                _logger.debug(
                    'Failed unitary condition, ||U^dU - I|| = %e' %
                    norm,
                )
            return False

        return True

    def __array__(
        self,
        dtype: np.typing.DTypeLike = np.complex128,
    ) -> npt.NDArray[np.complex128]:
        """Implements NumPy API for the UnitaryMatrix class."""
        if dtype != np.complex128:
            raise ValueError('UnitaryMatrix only supports Complex128 dtype.')

        return self._utry

    def __array_ufunc__(
        self,
        ufunc: np.ufunc,
        method: str,
        *inputs: npt.NDArray[Any],
        **kwargs: Any,
    ) -> UnitaryMatrix | npt.NDArray[np.complex128]:
        """Implements NumPy API for the UnitaryMatrix class."""
        if method != '__call__':
            return NotImplemented

        non_unitary_involved = False
        args: list[npt.NDArray[Any]] = []
        for input in inputs:
            if isinstance(input, UnitaryMatrix):
                args.append(input.numpy)
            else:
                args.append(input)
                non_unitary_involved = True

        out = ufunc(*args, **kwargs)

        # The results are unitary
        # if only unitaries are involved
        # and unitaries are closed under the specific operation.
        convert_back = (
            not non_unitary_involved and (
                ufunc.__name__ == 'conjugate'
                or ufunc.__name__ == 'matmul'
                or ufunc.__name__ == 'negative'
                or ufunc.__name__ == 'positive'
            )
            or (
                ufunc.__name__ == 'multiply'
                and all(
                    np.isscalar(input) or isinstance(input, UnitaryMatrix)
                    for input in inputs
                )
                and all(
                    np.abs(np.abs(input) - 1) <= 1e-14
                    for input in inputs if np.isscalar(input)
                )
            )
        )

        if convert_back:
            return UnitaryMatrix(out, self.radixes, False)

        return out

    def __str__(self) -> str:
        """Return the string representation of the unitary."""
        return str(self._utry)

    def __repr__(self) -> str:
        """Return the repr representation of the unitary."""
        return repr(self._utry)

    def __hash__(self) -> int:
        """Return the hash of the unitary."""
        return hash((self._utry[0][0], self._utry[-1][-1], self.shape))


UnitaryLike = Union[
    UnitaryMatrix,
    np.ndarray,
    Sequence[Sequence[Union[int, float, complex]]],
]



"""This module implements the Operation class."""
from __future__ import annotations

from typing import Any

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.ir.gates.composed.frozenparam import FrozenParameterGate
from bqskit.ir.location import CircuitLocation
from bqskit.ir.location import CircuitLocationLike
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_sequence


class Operation(DifferentiableUnitary):
    """An Operation groups together a gate, its parameters and location."""

    def __init__(
        self,
        gate: Gate,
        location: CircuitLocationLike,
        params: RealVector = [],
    ) -> None:
        """
        Construct an operation.

        Args:
            gate (Gate): The operation's gate.

            location (CircuitLocationLike):  The set of qudits this gate
                is applied to.

            params (RealVector): The parameters for the gate.

        Raises:
            ValueError: If `gate`'s size doesn't match `location`'s length.

            ValueError: If `gate`'s num_params doesn't match `params`'s
                length.
        """

        if not isinstance(gate, Gate):
            raise TypeError('Expected gate, got %s.' % type(gate))

        if not CircuitLocation.is_location(location):
            raise TypeError('Invalid location.')

        if is_sequence(params) and len(params) == 0 and gate.num_params != 0:
            params = [0.0] * gate.num_params

        gate.check_parameters(params)

        location = CircuitLocation(location)

        if len(location) != gate.num_qudits:
            raise ValueError('Gate and location size mismatch.')

        self._num_params = gate.num_params
        self._radixes = gate.radixes
        self._num_qudits = gate.num_qudits
        self._gate = gate
        self._location = location
        self._params = list(params)

    @property
    def gate(self) -> Gate:
        """The operation's gate."""
        return self._gate

    @property
    def location(self) -> CircuitLocation:
        """The qudit this operation is applied to."""
        return self._location

    @property
    def params(self) -> list[float]:
        """The operation's parameters for its gate."""
        return self._params

    @params.setter
    def params(self, params: list[float]) -> None:
        self.check_parameters(params)
        self._params = params

    def get_qasm(self) -> str:
        """
        Return the qasm string for this operation.

        Returns:
            str: The operation as a qasm line.
        """
        if isinstance(self.gate, FrozenParameterGate):
            full_params = self.gate.get_full_params(self.params)
        else:
            full_params = self.params

        return self.gate.get_qasm(full_params, self.location)

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        if len(params) != 0:
            return self.gate.get_unitary(params)

        return self.gate.get_unitary(self.params)

    def get_inverse(self) -> Operation:
        """Return the operation's inverse operation."""
        return Operation(
            self.gate.get_inverse(),
            self.location,
            self.gate.get_inverse_params(self.params),
        )

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this operation.

        See :class:`DifferentiableUnitary` for more info.
        """
        if len(params) != 0:
            return self.gate.get_grad(params)  # type: ignore

        return self.gate.get_grad(self.params)  # type: ignore

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        if len(params) != 0:
            return self.gate.get_unitary_and_grad(params)  # type: ignore

        return self.gate.get_unitary_and_grad(self.params)  # type: ignore

    def __eq__(self, rhs: Any) -> bool:
        """Check for equality."""
        if self is rhs:
            return True

        if not isinstance(rhs, Operation):
            return NotImplemented

        return (
            self.gate == rhs.gate
            and self.params == rhs.params
            and self.location == rhs.location
        )

    def __hash__(self) -> int:
        return hash((self.gate, self.location))

    def __str__(self) -> str:
        return f'{self.gate}@{self.location}'

    def __repr__(self) -> str:
        if len(self.params) == 0:
            return f'{self.gate}@{self.location}'
        return f'{self.gate}({self.params})@{self.location}'

    def is_differentiable(self) -> bool:
        """Check if operation is differentiable."""
        return isinstance(self.gate, DifferentiableUnitary)



"""This module implements the MessageDirection enum."""
from __future__ import annotations

from enum import IntEnum


class MessageDirection(IntEnum):
    """The direction a message came from."""

    BELOW = 0
    """This describes messages coming from a child manager/worker."""

    ABOVE = 1
    """This describes messages coming from a parent server/manager."""

    CLIENT = 2
    """This describes messages coming from a client."""

    SIGNAL = 3
    """This describes internal directives created by a signal handler."""
